"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Provides app settings dialog.
- Functionality: Builds UI elements based on a data map and interactively updates the dialog as settings are changed.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import Qt, QObject, QSize, QRegularExpression
from PySide6.QtWidgets import QDialog, QVBoxLayout, QTabWidget, QWidget, QSizePolicy, QPlainTextEdit, QScrollArea
from PySide6.QtWidgets import QLabel, QCheckBox, QLineEdit, QPushButton, QComboBox, QSpinBox, QDoubleSpinBox, QSlider

import logging
from typing import TYPE_CHECKING

from . import AppConfig
from . import Lexemes
from . import ThemeHelper

from ..enums.enum_base import EnumBase
from ..enums.languages import Languages
from ..enums.themes import Themes

from ..ui.enum_combo_box import EnumComboBox
from ..ui.horizontal_line_spacer import HorizontalLineSpacer
from ..ui.label_with_hint import LabelWithHint
from ..ui.dir_path_line_edit import DirPathLineEdit
from ..ui.file_path_line_edit import FilePathLineEdit

from ..modules.modules import Modules

if TYPE_CHECKING:
    from typing import Union  # noqa: F401


class SettingsDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent, Qt.WindowType.Window)

        self.setWindowState(Qt.WindowState.WindowActive)

        # Make the dialog non-modal
        self.setModal(False)

        self.parent = parent

        # Apply font from the dialog instance to the label
        self.setFont(self.parent.font())

        # self.settings = Settings(parent=self)
        # Use parent's settings to allow the Settings emitted signals to be caught
        self.settings = self.parent.settings  # type: ignore

        self.theme_helper = ThemeHelper()

        self.logger = logging.getLogger('settings_dialog')

        # Load lexemes for selected language and scope
        self.lexemes = Lexemes(self.settings.app_language, default_scope='settings_dialog')

        self.setWindowTitle(self.lexemes.get('window_title'))

        self.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Expanding)

        # Tabs widget as a main one
        self.tab_widget = None  # type: Union[QTabWidget, None]

        self.setStyleSheet(self.theme_helper.get_css('settings_dialog'))

        self.init_ui()

    def init_ui(self):
        # Dialog widget's main layout
        layout = QVBoxLayout(self)

        # Main tabs widget
        self.tab_widget = QTabWidget(self)
        self.tab_widget.setFont(self.font())
        self.tab_widget.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Expanding)
        # Add main tab widget to the layout
        layout.addWidget(self.tab_widget)

        # Get tabs fields config
        fields_conf = self.get_fields_conf()

        for conf in fields_conf:
            self.create_setting_field(conf)

        # Close button
        close_button = QPushButton(self.lexemes.get('button_close'))
        close_button.setFont(self.font())
        close_button.setObjectName('settings_dialog_button_close')
        close_button.clicked.connect(self.close)
        # Add close button to the layout
        layout.addWidget(close_button)

        self.connect_widgets()

        # Automatically adjust size to fit content
        # self.adjustSize()

        # Get the preferred size of the dialog's content
        preferred_size = self.parent.size()
        # Adjust the dialog's size based on the preferred size then
        if preferred_size.isValid():
            self.resize(QSize(int(preferred_size.width() * 0.5), int(preferred_size.height() * 0.9)))

    def get_fields_conf(self) -> list:
        # Load modules first to enable the loading of extension settings.
        module_instances = []
        for module in Modules().get_by_extension('settings_dialog'):
            # Pass settings object to avoid circular dependencies
            module_instances.append(Modules().create(module))

        fields_config = []
        fields_config.extend(self.get_general_fields())
        fields_config.extend(self.get_editor_config_fields())
        fields_config.extend(self.get_viewer_config_fields())
        fields_config.extend(self.get_ai_config_fields())

        # Extend settings UI based on extended settings
        for module_instance in module_instances:
            if hasattr(module_instance, 'extend_settings_dialog_fields_conf'):
                fields_config.extend(module_instance.extend_settings_dialog_fields_conf(self.tab_widget))
            else:
                self.logger.warning(f'Cannot extend settings_dialog for module {module_instance}')

        return fields_config

    def get_general_fields(self) -> list:
        # General
        tab_general = QWidget(self)

        # Create the scroll area
        scroll_area = QScrollArea(self)
        scroll_area.setObjectName('settings_dialog_tab_general')
        scroll_area.setWidgetResizable(True)

        # Layout for the General tab
        tab_general_layout = QVBoxLayout(tab_general)

        # Set the content widget inside the scroll area
        scroll_area.setWidget(tab_general)

        self.tab_widget.addTab(scroll_area, self.lexemes.get('tab_general'))

        return [
            # [General]
            # General settings block label
            {"type": QLabel, "name": "settings_dialog_general_app_config_label",
             "props": {"setProperty": ("class", "group-header-label")},
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('general_app_config_label'), "style": {"bold": True},
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Available languages label
            {"type": QLabel, "name": "settings_dialog_general_app_language_label",
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('general_app_language_label'),
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Available languages dropdown list
            {"type": EnumComboBox,
             "args": [sorted(Languages, key=lambda member: (not member.is_default, str(member.value)))],
             "name": "settings_dialog_general_app_language_combo:app_language",  # Lexeme key : Object name
             "callback": lambda obj: tab_general_layout.addWidget(obj),
             "placeholder_text": self.lexemes.get('general_app_language_combo_placeholder_text'),
             "accessible_description":
                 self.lexemes.get('app_language_combo_accessible_description')},
            # Available themes label
            {"type": QLabel, "name": "settings_dialog_general_app_theme_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('general_app_theme_label'),
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Available themes dropdown list
            {"type": EnumComboBox, "args": [sorted(Themes, key=lambda member: (not member.is_default, str(member.value)))],
             "name": "settings_dialog_general_app_theme_combo:app_theme",  # Lexeme key : Object name
             "callback": lambda obj: tab_general_layout.addWidget(obj),
             "placeholder_text": self.lexemes.get('general_app_theme_combo_placeholder_text'),
             "accessible_description":
                 self.lexemes.get('general_app_theme_combo_accessible_description')},
            # Horizontal spacer
            {"type": HorizontalLineSpacer, "callback": lambda obj: tab_general_layout.addWidget(obj)},
            # Main menu label
            {"type": QLabel, "name": "settings_dialog_general_app_main_menu_label",
             "props": {"setProperty": ("class", "group-header-label")},
             "alignment": Qt.AlignmentFlag.AlignLeft, "style": {"bold": True},
             "text": self.lexemes.get('general_app_main_menu_label'),
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Either show or not main menu
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_general_app_main_menu_checkbox:show_main_menu",
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('general_app_main_menu_checkbox'),
             "accessible_description":
                 self.lexemes.get('general_app_main_menu_checkbox_accessible_description')},
            # Horizontal spacer
            {"type": HorizontalLineSpacer, "callback": lambda obj: tab_general_layout.addWidget(obj)},
            # Status bar settings block label
            {"type": QLabel, "name": "settings_dialog_general_statusbar_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "props": {"setProperty": ("class", "group-header-label")},
             "text": self.lexemes.get('general_statusbar_label'), "style": {"bold": True},
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Either show or not global position at status bar
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_general_statusbar_show_global_cursor_position_checkbox"
                     ":show_global_cursor_position",
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('general_statusbar_show_global_cursor_position_checkbox'),
             "accessible_description":
                 self.lexemes.get('general_statusbar_show_global_cursor_position_checkbox_accessible_description')},
            # Horizontal spacer
            {"type": HorizontalLineSpacer, "callback": lambda obj: tab_general_layout.addWidget(obj)},
            # Main menu label
            {"type": QLabel, "name": "settings_dialog_general_app_font_size_label",
             "props": {"setProperty": ("class", "group-header-label")},
             "alignment": Qt.AlignmentFlag.AlignLeft, "style": {"bold": True},
             "text": self.lexemes.get('general_app_font_size_label', size=self.settings.app_font_size),
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            {"type": QSlider, "args": [Qt.Orientation.Horizontal],
             "props": {'setFocusPolicy': Qt.FocusPolicy.StrongFocus, 'setTickPosition': QSlider.TickPosition.TicksAbove,
                       'setTickInterval': 5, 'setSingleStep': 1,
                       'setMinimum': AppConfig().get_font_min_size(),
                       'setMaximum': AppConfig().get_font_max_size()},
             "name": "settings_dialog_general_app_font_size_slider:app_font_size",  # Lexeme key : Object name
             "callback": lambda obj: tab_general_layout.addWidget(obj),
             "accessible_description":
                 self.lexemes.get('general_app_font_size_slider_accessible_description')},
            # Spacer to keep elements above on top
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding),
             "callback": lambda obj: tab_general_layout.addWidget(obj)},
        ]

    def get_editor_config_fields(self) -> list:
        # Editor
        tab_editor_config = QWidget(self)

        # Create the scroll area
        scroll_area = QScrollArea(self)
        scroll_area.setObjectName('settings_dialog_tab_editor_config')
        scroll_area.setWidgetResizable(True)

        # Layout for the Editor Config tab
        tab_editor_config_layout = QVBoxLayout(tab_editor_config)

        # Set the content widget inside the scroll area
        scroll_area.setWidget(tab_editor_config)

        self.tab_widget.addTab(scroll_area, self.lexemes.get('tab_editor_config'))

        return [
            # [Editor config]
            # Editor block label
            {"type": QLabel, "name": "settings_dialog_editor_config_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "props": {"setProperty": ("class", "group-header-label")},
             "text": self.lexemes.get('editor_config_label'), "style": {"bold": True},
             "callback": lambda obj: tab_editor_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Either to show or not editor's line numbers
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_editor_config_show_line_numbers_checkbox:show_line_numbers",
             "callback": lambda obj: tab_editor_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('editor_config_show_line_numbers_checkbox'),
             "accessible_description":
                 self.lexemes.get('editor_config_show_line_numbers_checkbox_accessible_description')},
            # Spacer to keep elements above on top
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding),
             "callback": lambda obj: tab_editor_config_layout.addWidget(obj)},
        ]

    def get_viewer_config_fields(self) -> list:
        # Viewer
        tab_viewer_config = QWidget(self)

        # Create the scroll area
        scroll_area = QScrollArea(self)
        scroll_area.setObjectName('settings_dialog_tab_viewer_config')
        scroll_area.setWidgetResizable(True)

        # Layout for the Viewer Config tab
        tab_viewer_config_layout = QVBoxLayout(tab_viewer_config)

        # Set the content widget inside the scroll area
        scroll_area.setWidget(tab_viewer_config)

        self.tab_widget.addTab(scroll_area, self.lexemes.get('tab_viewer_config'))

        return [
            # [Viewer config]
            # Viewer block label
            {"type": QLabel, "name": "settings_dialog_viewer_config_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "props": {"setProperty": ("class", "group-header-label")},
             "text": self.lexemes.get('viewer_config_label'), "style": {"bold": True},
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Either to show or not viewer's emojis
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_viewer_config_process_emojis_checkbox:viewer_process_emojis",
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('viewer_config_process_emojis_checkbox'),
             "accessible_description":
                 self.lexemes.get('viewer_config_process_emojis_checkbox_accessible_description')},
            # Spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj)},
            # Either to show or not viewer's TODOs
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_viewer_config_highlight_todos_checkbox:viewer_highlight_todos",
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('viewer_config_highlight_todos_checkbox'),
             "accessible_description":
                 self.lexemes.get('viewer_config_highlight_todos_checkbox_accessible_description')},
            # Spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj)},
            # Either to show open link confirmation dialog or not
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_viewer_config_open_link_confirmation_checkbox:viewer_open_link_confirmation",
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('viewer_config_open_link_confirmation_checkbox'),
             "accessible_description":
                 self.lexemes.get('viewer_config_open_link_confirmation_checkbox_accessible_description')},
            # Spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj)},
            # Auto save downloaded resources on disk
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_viewer_config_save_resources_checkbox:viewer_save_resources",
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('viewer_config_save_resources_checkbox'),
             "accessible_description":
                 self.lexemes.get('viewer_config_save_resources_checkbox_accessible_description')},
            # Spacer to keep elements above on top
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding),
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj)},
        ]

    def get_ai_config_fields(self) -> list:
        # AI Config
        tab_ai_config = QWidget(self)

        # Create the scroll area
        scroll_area = QScrollArea(self)
        scroll_area.setObjectName('settings_dialog_tab_ai_config')
        scroll_area.setWidgetResizable(True)

        # Layout for the AI Config tab
        tab_ai_config_layout = QVBoxLayout(tab_ai_config)

        # Set the content widget inside the scroll area
        scroll_area.setWidget(tab_ai_config)

        self.tab_widget.addTab(scroll_area, self.lexemes.get('tab_ai_config'))

        return [
            # [AI config]
            # Default inference model label
            {"type": QLabel, "name": "settings_dialog_ai_config_inference_module_label",
             "props": {"setProperty": ("class", "group-header-label")},
             "alignment": Qt.AlignmentFlag.AlignLeft, "style": {"bold": True},
             "text": self.lexemes.get('ai_config_inference_module_label'),
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Supported models label
            {"type": LabelWithHint, "kwargs": {
                "tooltip": ('ai_config_inference_module_names_combo_accessible_description',
                            self.lexemes.get('ai_config_inference_module_names_combo_accessible_description'))},
             "name": "settings_dialog_ai_config_inference_module_names_combo_label",
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('ai_config_inference_module_names_combo_label'),
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Supported models dropdown list
            {"type": EnumComboBox,
             "args": [sorted(EnumBase('InferenceModuleNames', self.settings.ai_config_inference_modules),
                      key=lambda member: (not member.is_default, str(member.value)))],
             "name": "settings_dialog_ai_config_inference_module_names_combo:ai_config_inference_module",
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj),
             "placeholder_text": self.lexemes.get('ai_config_inference_module_names_combo_placeholder_text'),
             "accessible_description":
                 self.lexemes.get('ai_config_inference_module_names_combo_accessible_description')},
            # Horizontal spacer
            {"type": HorizontalLineSpacer, "callback": lambda obj: tab_ai_config_layout.addWidget(obj)},
            # Base AI settings block label
            {"type": QLabel, "name": "settings_dialog_ai_config_base_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "props": {"setProperty": ("class", "group-header-label")},
             "text": self.lexemes.get('ai_config_base_label'), "style": {"bold": True},
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Auto save downloaded resources on disk
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_ai_config_multi_turn_dialogue_checkbox:ai_config_multi_turn_dialogue",
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('ai_config_multi_turn_dialogue_checkbox'),
             "accessible_description":
                 self.lexemes.get('ai_config_multi_turn_dialogue_checkbox_accessible_description')},
            # Spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj)},
            # Auto save downloaded resources on disk
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_ai_config_convert_to_md_checkbox:ai_config_convert_to_md",
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('ai_config_convert_to_md_checkbox'),
             "accessible_description": self.lexemes.get('ai_config_convert_to_md_checkbox_accessible_description')},
            # Spacer to keep elements above on top
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding),
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj)},
        ]

    def apply_props(self, widget, properties):
        for prop, value in properties.items():
            # Get the method based on the property name
            setter = getattr(widget, prop, None)
            if callable(setter):
                if type(value) is tuple:
                    setter(*value)  # Call the setter with the provided values
                else:
                    setter(value)  # Call the setter with the provided value

    def create_setting_field(self, conf) -> QWidget:  # noqa: C901
        """
        Create a UI setting field dynamically based on the provided configuration.

        Args:
            conf (dict): Configuration for the field. Includes type, properties, callbacks, and visual settings.

        Returns:
            QWidget: The dynamically created UI object.
        """

        # Extract arguments for the object initialization
        _args = conf.get('args', [self])  # Default to adding to the dialog
        _kwargs = conf.get('kwargs', {})

        # Instantiate the UI object of the specified type
        obj = conf['type'](*_args, **_kwargs)

        # Apply dialog-wide font if supported (e.g., QLabel, QLineEdit, etc.)
        if hasattr(obj, 'setFont'):
            obj.setFont(self.font())

        # Apply custom properties (if defined in the configuration)
        if 'props' in conf:
            self.apply_props(obj, conf['props'])

        # Set object name for better debugging and callback management
        if 'name' in conf and conf['name']:
            obj.setObjectName(conf['name'])

        # Apply common text-related properties
        for key, method in [
            ('text', 'setText'),
            ('read_only', 'setReadOnly'),
            ('max_length', 'setMaxLength'),
            ('placeholder_text', 'setPlaceholderText'),
            ('accessible_description', 'setAccessibleDescription'),
        ]:
            if key in conf and hasattr(obj, method):
                getattr(obj, method)(conf[key])

        # Apply size policies if specified
        if 'size_policy' in conf:
            params = conf['size_policy']
            if isinstance(params, tuple):
                obj.setSizePolicy(*params)
            else:
                obj.setSizePolicy(params)

        # Set alignment if supported and specified
        if 'alignment' in conf and hasattr(obj, 'setAlignment'):
            obj.setAlignment(conf['alignment'])

        # Handle layout-related callbacks
        if 'callback' in conf and callable(conf['callback']):
            conf['callback'](obj)

        # Handle value change callbacks for interactive widgets
        if 'on_value_change' in conf and callable(conf['on_value_change']):
            obj.valueChanged.connect(lambda: conf['on_value_change'](source_object=obj, source_widget=self))

        # Apply styles (bold, italic, color) via stylesheet
        if 'style' in conf:
            styles = []
            if 'bold' in conf['style']:
                styles.append('font-weight: bold;')
            if 'italic' in conf['style']:
                styles.append('font-style: italic;')
            if 'color' in conf['style']:
                styles.append(f"color: {conf['style']['color']};")
            if styles:
                obj.setStyleSheet(' '.join(styles))

        # Set the height for QPlainTextEdit based on the number of text lines (if specified)
        if 'text_lines' in conf and isinstance(obj, QPlainTextEdit) and int(conf['text_lines']) > 0:
            font_metrics = obj.fontMetrics()
            lines_height = font_metrics.height() * conf['text_lines']
            obj.setFixedHeight(lines_height)

        # Return the dynamically created object
        return obj

    def parse_object_name(self, object_name: str):
        return self.settings.settings_helper.parse_object_name(object_name)

    def connect_widgets(self):  # noqa: C901 - consider simplifying this method
        # Find children of type QCheckBox
        checkboxes = self.findChildren(QCheckBox)
        for checkbox in checkboxes:
            if isinstance(checkbox, QCheckBox):
                # Parse the object name in case it contains a combination of lexeme and setting keys
                _lexeme_key, setting_name = self.parse_object_name(checkbox.objectName())
                if hasattr(self.settings, setting_name):
                    checkbox.setChecked(getattr(self.settings, setting_name, False))
                # Connect signal after set up defaults or restore saved value to avoid signal emitting right after.
                checkbox.stateChanged.connect(self.save_settings)

        # Find children of type QComboBox
        combo_boxes = self.findChildren(QComboBox)
        for combo_box in combo_boxes:
            if isinstance(combo_box, EnumComboBox):
                # Parse the object name in case it contains a combination of lexeme and setting keys
                _lexeme_key, setting_name = self.parse_object_name(combo_box.objectName())
                if hasattr(self.settings, setting_name):
                    setting_value = getattr(self.settings, setting_name, None)
                    # Get the index of the Enum member by the name stored in the settings (e.g. Language's 'EN')
                    index = next((i for i, val in enumerate(combo_box.enum_class) if val.name.lower() == setting_value),
                                 None)
                    self.logger.debug(f'Enum {combo_box.enum_class} index for value: {index} [{setting_value}]')
                    if index is not None:
                        combo_box.setCurrentIndex(index)
                # Connect signal after set up defaults or restore saved value to avoid signal emitting right after.
                combo_box.currentIndexChanged.connect(self.save_settings)

        # Find children of type QSpinBox
        spin_boxes = self.findChildren(QSpinBox)
        for spin_box in spin_boxes:
            if isinstance(spin_box, QSpinBox):
                # Parse the object name in case it contains a combination of lexeme and setting keys
                _lexeme_key, setting_name = self.parse_object_name(spin_box.objectName())
                if hasattr(self.settings, setting_name):
                    spin_box.setValue(getattr(self.settings, setting_name, 0))
                spin_box.valueChanged.connect(self.save_settings)

        # Find children of type QDoubleSpinBox
        double_spin_boxes = self.findChildren(QDoubleSpinBox)
        for double_spin_box in double_spin_boxes:
            if isinstance(double_spin_box, QDoubleSpinBox):
                # Parse the object name in case it contains a combination of lexeme and setting keys
                _lexeme_key, setting_name = self.parse_object_name(double_spin_box.objectName())
                if hasattr(self.settings, setting_name):
                    double_spin_box.setValue(getattr(self.settings, setting_name, 0))
                double_spin_box.valueChanged.connect(self.save_settings)

        # Find children of type QLineEdit
        line_edits = self.findChildren(QLineEdit)
        for line_edit in line_edits:
            if isinstance(line_edit, QLineEdit):
                # Parse the object name in case it contains a combination of lexeme and setting keys
                _lexeme_key, setting_name = self.parse_object_name(line_edit.objectName())
                if hasattr(self.settings, setting_name):
                    line_edit.setText(str(getattr(self.settings, setting_name, '')))
                # Connect signal after set up defaults or restore saved value to avoid signal emitting right after.
                line_edit.textChanged.connect(self.save_settings)

        # Find children of type QPlainTextEdit
        text_edits = self.findChildren(QPlainTextEdit)
        for text_edit in text_edits:
            if isinstance(text_edit, QPlainTextEdit):
                # Parse the object name in case it contains a combination of lexeme and setting keys
                _lexeme_key, setting_name = self.parse_object_name(text_edit.objectName())
                if hasattr(self.settings, setting_name):
                    text_edit.setPlainText(getattr(self.settings, setting_name, ''))
                # Connect signal after set up defaults or restore saved value to avoid signal emitting right after.
                text_edit.textChanged.connect(self.save_settings)

        # Find children of type QSlider
        sliders = self.findChildren(QSlider)
        for slider in sliders:
            if isinstance(slider, QSlider):
                # Parse the object name in case it contains a combination of lexeme and setting keys
                _lexeme_key, setting_name = self.parse_object_name(slider.objectName())
                if hasattr(self.settings, setting_name):
                    slider.setValue(getattr(self.settings, setting_name, 0))
                slider.valueChanged.connect(self.save_settings)

    def save_settings(self):  # noqa: C901 - consider simplifying this method
        # Determine which widget emitted the signal by the object name set
        sender_widget = self.sender()
        sender_name = sender_widget.objectName()

        if sender_name.startswith('qt_'):
            # For QSpinBox or QDoubleSpinBox, 'qt_spinbox_lineedit' refers to the internal QLineEdit widget
            # used for handling text input (e.g., via typing).
            if sender_widget and sender_name == "qt_spinbox_lineedit":
                sender_widget = sender_widget.parent()  # Resolve to the parent widget (QSpinBox).
                sender_name = sender_widget.objectName()
            else:
                self.logger.warning(f"Unhandled internal widget type: {sender_name}")
                return

        # Parse the object name in case it contains a combination of lexeme and setting keys
        lexeme_key, setting_name = self.parse_object_name(sender_name)

        setting_value = None
        setting_text = None
        if isinstance(sender_widget, QCheckBox):
            setting_value = sender_widget.isChecked()
            setting_text = sender_widget.text()
        elif isinstance(sender_widget, QComboBox):
            # Index of the item in the enum
            # index = sender_widget.currentIndex()
            # setting_value = sender_widget.currentData()
            setting_text = sender_widget.currentText()
            """
            To get enum name instead of the value, say for Languages:
            index = sender_widget.currentIndex()
            setting_value = list(Languages)[index].name
            """
            setting_value = sender_widget.currentData().name.lower()  # Save in lower case
        elif isinstance(sender_widget, QSpinBox):
            setting_value = sender_widget.value()
            setting_text = sender_widget.text()
        elif isinstance(sender_widget, QDoubleSpinBox):
            setting_value = sender_widget.value()
            setting_text = sender_widget.text()
        elif isinstance(sender_widget, QLineEdit):
            setting_value = sender_widget.text()
            setting_text = sender_widget.placeholderText()
        elif isinstance(sender_widget, QPlainTextEdit):
            setting_value = sender_widget.toPlainText()
            setting_text = sender_widget.placeholderText()
        elif isinstance(sender_widget, QSlider):
            setting_value = sender_widget.value()
            setting_text = sender_widget.accessibleDescription()

        self.logger.debug(f"Saving setting '{setting_name}': {setting_value} ({setting_text})")

        try:
            setattr(self.settings, setting_name, setting_value)
        except AttributeError as e:
            self.logger.warning(f'ERROR: {e}')

        self.logger.debug(f"Setting new value: {getattr(self.settings, setting_name)}")

        """
        In case of language change try to update as much text labels as possible.
        It may not cover all the labels, but may help with update anyway.
        """
        if setting_name == 'app_language' or sender_widget.objectName().endswith('app_language'):
            # Update lexemes object as app language has just been changed
            self.lexemes = Lexemes(self.settings.app_language, default_scope='settings_dialog')
            # Get all the new lexemes
            app_language_lexemes = self.lexemes.get_all()

            # Load modules to retrieve lexemes for the settings dialog
            for module in Modules().get_by_extension('settings_dialog'):
                # Pass settings object to avoid circular dependencies
                module_instance = Modules().create(module)
                if hasattr(module_instance, 'lexemes'):
                    app_language_lexemes['settings_dialog'].update(
                        module_instance.lexemes.get_by_scope('settings_dialog'))

            # Update dialog title at first
            self.setWindowTitle(self.lexemes.get('window_title'))
            # Iterate each lexeme's scope to match it with object name to update the text if applicable
            for scope in app_language_lexemes:
                self.logger.debug(f'Lexeme scope to check updates {scope}')
                # Iterate each lexeme from the scope
                for lexeme_key in app_language_lexemes[scope].keys():
                    """
                    Some objects may have name set with ":" delimiter, which means "lexeme_key:setting_name".
                    It may help to transfer more params with object name with ease.
                    """
                    # Parse the object name in case it contains a combination of lexeme and setting keys
                    lexeme_key, setting_name = self.parse_object_name(lexeme_key)

                    # Get lexeme in the new language set
                    lexeme = app_language_lexemes[scope][lexeme_key]

                    """
                    Search with regex as the object name should contain scope and may contain setting name as well,
                    but the lexeme key is a crucial part of the name in this search.
                    """
                    regex = QRegularExpression(f"(?={scope}_)?{lexeme_key}.*?")
                    # The self.findChildren() method results contain only the elements related to this dialog
                    found_objects = self.findChildren(QObject, regex, Qt.FindChildOption.FindChildrenRecursively)
                    if not found_objects:
                        # Try to find it in the parent dialog to update the main window elements
                        found_objects = self.parent.findChildren(QObject, regex,
                                                                 Qt.FindChildOption.FindChildrenRecursively)
                        if found_objects:
                            self.logger.debug(f"Found object to update ({scope}){lexeme_key}: {found_objects}")

                    # Search by the full object name
                    # object_name = self.lexemes.get_full_key(lexeme_key)
                    # found_objects = self.findChildren(QObject, object_name, Qt.FindChildrenRecursively)

                    for obj in found_objects:
                        if (isinstance(obj, (QLabel, LabelWithHint, QCheckBox, QComboBox, QPushButton))
                                and hasattr(obj, 'setText')
                                and callable(getattr(obj, 'setText'))):
                            self.logger.debug(
                                f'Object lexeme update: {obj.objectName()} with id: {id(obj)}, lexeme "{lexeme}"')
                            obj.setText(lexeme)
                        if isinstance(obj, LabelWithHint):
                            if obj.property('tooltip_lexeme'):
                                # Lexeme must be default_scope='settings_dialog'
                                obj.set_tooltip(self.lexemes.get(obj.property('tooltip_lexeme')))
                        if isinstance(obj, QScrollArea):
                            # Add iterable map of the tabs to change the search by QWidget to explicitly set tab text.
                            self.set_tab_text(obj.objectName(), lexeme)
                        # Exclusions
                        if obj.objectName() == "settings_dialog_general_app_font_size_label":
                            obj.setText(
                                self.lexemes.get('general_app_font_size_label', size=self.settings.app_font_size))

        if setting_name == 'app_theme' or sender_widget.objectName().endswith('app_theme'):
            # Update the widget's stylesheet with the selected theme
            self.setStyleSheet(self.theme_helper.get_css('settings_dialog'))
            # Update font size to correct the tab widget's font
            self.update_font_size(font_size=self.settings.app_font_size)

            # Search objects to update by their type
            found_objects = self.findChildren(LabelWithHint)
            found_objects += self.findChildren(DirPathLineEdit)
            found_objects += self.findChildren(FilePathLineEdit)
            for obj in found_objects:
                if (isinstance(obj, (LabelWithHint, DirPathLineEdit, FilePathLineEdit))
                        and hasattr(obj, 'load_icon')
                        and callable(getattr(obj, 'load_icon'))):
                    obj.load_icon()

        if setting_name == 'app_font_size' or sender_widget.objectName().endswith('app_font_size'):
            # Update the font size of elements
            self.update_font_size(font_size=setting_value)

    def update_font_size(self, font_size: int):
        # Update the font from the parent to ensure it reflects the new font size.
        # A simple font size update should also work.
        self.setFont(self.parent.font())
        widgets_to_update = [QLabel, QTabWidget, QPushButton, QCheckBox, QLineEdit, QPlainTextEdit, QComboBox,
                             QSpinBox, QDoubleSpinBox, QSlider]
        for _widget in widgets_to_update:
            # Find all QLabel objects
            found_objects = self.findChildren(_widget)
            for obj in found_objects:
                # Align with the dialog font size (mostly for QLabel)
                if hasattr(obj, 'setFont'):
                    obj.setFont(self.font())
                if hasattr(obj, 'toolTip'):
                    obj.setStyleSheet("QToolTip { font-size: %dpt; }" % font_size)
                if obj.objectName() == "settings_dialog_general_app_font_size_label":
                    """
                    # Update the font size on the label
                    font = QFont()
                    font.setPointSize(setting_value)
                    app_font_size_label.setFont(font)
                    """
                    obj.setText(self.lexemes.get('general_app_font_size_label', size=font_size))

    def set_tab_text(self, tab_object_name, text):
        # Iterate through all the tabs
        for index in range(self.tab_widget.count()):
            # Get the widget for the current index
            widget = self.tab_widget.widget(index)  # type: Union[QScrollArea, QWidget]

            # Check if this widget matches the object name
            if widget.objectName() == tab_object_name:
                # Set the text
                self.tab_widget.setTabText(index, text)
                break



# Russian lexemes settings_dialog.py
lexemes = {
    # Настройки
    "window_title": "Настройки",

    "button_close": "Закрыть",

    "tab_general": "Общие",
    "tab_editor_config": "Редактор",
    "tab_viewer_config": "Просмотрщик",
    "tab_ai_config": "Настройки ИИ",

    "general_app_config_label": "Настройки приложения",
    "general_app_language_label": "Язык",
    "general_app_language_combo_placeholder_text": "Выбрать язык",
    "general_app_language_combo_accessible_description": "Язык интерфейса приложения",
    "general_app_theme_label": "Тема",
    "general_app_theme_combo_placeholder_text": "Выбрать тему",
    "general_app_theme_combo_accessible_description": "Тема интерфейса приложения",
    "general_app_main_menu_label": "Главное меню",
    "general_app_main_menu_checkbox": "Показать главное меню",
    "general_app_main_menu_checkbox_accessible_description": "Показать основное выпадающее меню приложения",
    "general_app_font_size_label": "Размер шрифта: {size}pt",
    "general_app_font_size_slider_accessible_description": "Настроить глобальный размер шрифта приложения",

    "general_statusbar_label": "Строка состояния",
    "general_statusbar_show_global_cursor_position_checkbox": "Показать глобальное положение курсора",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Отображать глобальное положение курсора в строке состояния",

    "editor_config_label": "Настройки редактора",
    "editor_config_show_line_numbers_checkbox": "Показать номера строк",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Отображать номера строк в редакторе",

    "viewer_config_label": "Настройки просмотрщика",
    "viewer_config_process_emojis_checkbox": "Преобразовать текстовые эмодзи в графику",
    "viewer_config_process_emojis_checkbox_accessible_description":
        "Преобразовать текстовые эмодзи в графические изображения",
    "viewer_config_highlight_todos_checkbox": "Выделить задачи TODO",
    "viewer_config_highlight_todos_checkbox_accessible_description": "Выделять задачи TODO в тексте",
    "viewer_config_open_link_confirmation_checkbox": "Требовать подтверждение перед открытием ссылок",
    "viewer_config_open_link_confirmation_checkbox_accessible_description":
        "Запрашивать подтверждение перед открытием ссылок",
    "viewer_config_save_resources_checkbox": "Автосохранение внешних изображений на диск",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Автоматически сохраняет копии внешних изображений на диск для доступа без подключения к интернету.",

    "ai_config_inference_module_label": "Модуль вывода",
    "ai_config_inference_module_names_combo_label": "Активный модуль вывода",
    "ai_config_inference_module_names_combo_placeholder_text": "Выбрать модуль",
    "ai_config_inference_module_names_combo_accessible_description":
        "Выберите из доступных модулей ИИ-инференции для работы с AI Assistant. Варианты включают локальные\n"
        "большие языковые модели (LLM) с обработкой в реальном времени или функциональность на базе API.",

    "ai_config_base_label": "Базовые параметры",
    "ai_config_multi_turn_dialogue_checkbox": "Многошаговый чат-диалог с памятью",
    "ai_config_multi_turn_dialogue_checkbox_accessible_description":
        "Включите многошаговый чат-диалог, который сохраняет последний запрос для памяти.\n"
        "Когда выключено, только новое сообщение и системный промпт влияют на ответ.",
    "ai_config_convert_to_md_checkbox": "Конвертировать результат в Markdown",
    "ai_config_convert_to_md_checkbox_accessible_description":
        "Преобразуйте выходное сообщение в формат Markdown.",
}



# Spanish lexemes color_picker_dialog.py
lexemes = {
    "color_picker_title": "Seleccionar un color",

    "color_picker_color_black": "Negro",
    "color_picker_color_silver": "Plateado",
    "color_picker_color_gray": "Gris",
    "color_picker_color_white": "Blanco",
    "color_picker_color_maroon": "Marrón",
    "color_picker_color_red": "Rojo",
    "color_picker_color_purple": "Púrpura",
    "color_picker_color_fuchsia": "Fucsia",
    "color_picker_color_green": "Verde",
    "color_picker_color_lime": "Lima",
    "color_picker_color_olive": "Oliva",
    "color_picker_color_yellow": "Amarillo",
    "color_picker_color_navy": "Azul marino",
    "color_picker_color_blue": "Azul",
    "color_picker_color_teal": "Verde azulado",
    "color_picker_color_aqua": "Aguamarina",
    "color_picker_color_aliceblue": "Azul Alicia",
    "color_picker_color_antiquewhite": "Blanco antiguo",
    "color_picker_color_aquamarine": "Aguamarina",
    "color_picker_color_azure": "Azur",
    "color_picker_color_beige": "Beige",
    "color_picker_color_bisque": "Bisque",
    "color_picker_color_blanchedalmond": "Almendra blanqueada",
    "color_picker_color_blueviolet": "Violeta azulado",
    "color_picker_color_brown": "Marrón",
    "color_picker_color_burlywood": "Madera",
    "color_picker_color_cadetblue": "Azul cadete",
    "color_picker_color_chartreuse": "Chartreuse",
    "color_picker_color_chocolate": "Chocolate",
    "color_picker_color_coral": "Coral",
    "color_picker_color_cornflowerblue": "Azul aciano",
    "color_picker_color_cornsilk": "Seda de maíz",
    "color_picker_color_crimson": "Carmesí",
    "color_picker_color_cyan": "Cian",
    "color_picker_color_darkblue": "Azul oscuro",
    "color_picker_color_darkcyan": "Cian oscuro",
    "color_picker_color_darkgoldenrod": "Vara de oro oscuro",
    "color_picker_color_darkgray": "Gris oscuro",
    "color_picker_color_darkgreen": "Verde oscuro",
    "color_picker_color_darkkhaki": "Caqui oscuro",
    "color_picker_color_darkmagenta": "Magenta oscuro",
    "color_picker_color_darkolivegreen": "Verde oliva oscuro",
    "color_picker_color_darkorange": "Naranja oscuro",
    "color_picker_color_darkorchid": "Orquídea oscura",
    "color_picker_color_darkred": "Rojo oscuro",
    "color_picker_color_darksalmon": "Salmón oscuro",
    "color_picker_color_darkseagreen": "Verde mar oscuro",
    "color_picker_color_darkslateblue": "Azul pizarra oscuro",
    "color_picker_color_darkslategray": "Gris pizarra oscuro",
    "color_picker_color_darkturquoise": "Turquesa oscuro",
    "color_picker_color_darkviolet": "Violeta oscuro",
    "color_picker_color_deeppink": "Rosa profundo",
    "color_picker_color_deepskyblue": "Azul cielo profundo",
    "color_picker_color_dimgray": "Gris tenue",
    "color_picker_color_dodgerblue": "Azul dodger",
    "color_picker_color_firebrick": "Ladrillo refractario",
    "color_picker_color_floralwhite": "Blanco floral",
    "color_picker_color_forestgreen": "Verde bosque",
    "color_picker_color_gainsboro": "Gainsboro",
    "color_picker_color_ghostwhite": "Blanco fantasma",
    "color_picker_color_gold": "Oro",
    "color_picker_color_goldenrod": "Vara de oro",
    "color_picker_color_greenyellow": "Verde amarillo",
    "color_picker_color_honeydew": "Melón",
    "color_picker_color_hotpink": "Rosa intenso",
    "color_picker_color_indianred": "Rojo indio",
    "color_picker_color_indigo": "Índigo",
    "color_picker_color_ivory": "Marfil",
    "color_picker_color_khaki": "Caqui",
    "color_picker_color_lavender": "Lavanda",
    "color_picker_color_lavenderblush": "Rubor lavanda",
    "color_picker_color_lawngreen": "Verde césped",
    "color_picker_color_lemonchiffon": "Chifón limón",
    "color_picker_color_lightblue": "Azul claro",
    "color_picker_color_lightcoral": "Coral claro",
    "color_picker_color_lightcyan": "Cian claro",
    "color_picker_color_lightgoldenrodyellow": "Amarillo vara de oro claro",
    "color_picker_color_lightgray": "Gris claro",
    "color_picker_color_lightgreen": "Verde claro",
    "color_picker_color_lightpink": "Rosa claro",
    "color_picker_color_lightsalmon": "Salmón claro",
    "color_picker_color_lightseagreen": "Verde mar claro",
    "color_picker_color_lightskyblue": "Azul cielo claro",
    "color_picker_color_lightslategray": "Gris pizarra claro",
    "color_picker_color_lightsteelblue": "Azul acero claro",
    "color_picker_color_lightyellow": "Amarillo claro",
    "color_picker_color_limegreen": "Verde lima",
    "color_picker_color_linen": "Lino",
    "color_picker_color_magenta": "Magenta",
    "color_picker_color_mediumaquamarine": "Aguamarina medio",
    "color_picker_color_mediumblue": "Azul medio",
    "color_picker_color_mediumorchid": "Orquídea media",
    "color_picker_color_mediumpurple": "Púrpura medio",
    "color_picker_color_mediumseagreen": "Verde mar medio",
    "color_picker_color_mediumslateblue": "Azul pizarra medio",
    "color_picker_color_mediumspringgreen": "Verde primavera medio",
    "color_picker_color_mediumturquoise": "Turquesa medio",
    "color_picker_color_mediumvioletred": "Rojo violeta medio",
    "color_picker_color_midnightblue": "Azul medianoche",
    "color_picker_color_mintcream": "Crema de menta",
    "color_picker_color_mistyrose": "Rosa brumoso",
    "color_picker_color_moccasin": "Mocasín",
    "color_picker_color_navajowhite": "Blanco Navajo",
    "color_picker_color_oldlace": "Encaje antiguo",
    "color_picker_color_olivedrab": "Oliva desvaído",
    "color_picker_color_orange": "Naranja",
    "color_picker_color_orangered": "Rojo naranja",
    "color_picker_color_orchid": "Orquídea",
    "color_picker_color_palegoldenrod": "Vara de oro pálido",
    "color_picker_color_palegreen": "Verde pálido",
    "color_picker_color_paleturquoise": "Turquesa pálido",
    "color_picker_color_palevioletred": "Rojo violeta pálido",
    "color_picker_color_papayawhip": "Papaya batida",
    "color_picker_color_peachpuff": "Bofetada de durazno",
    "color_picker_color_peru": "Perú",
    "color_picker_color_pink": "Rosa",
    "color_picker_color_plum": "Ciruela",
    "color_picker_color_powderblue": "Azul talco",
    "color_picker_color_rebeccapurple": "Púrpura Rebecca",
    "color_picker_color_rosybrown": "Marrón rosado",
    "color_picker_color_royalblue": "Azul real",
    "color_picker_color_saddlebrown": "Marrón silla",
    "color_picker_color_salmon": "Salmón",
    "color_picker_color_sandybrown": "Marrón arenoso",
    "color_picker_color_seagreen": "Verde mar",
    "color_picker_color_seashell": "Concha",
    "color_picker_color_sienna": "Siena",
    "color_picker_color_skyblue": "Azul cielo",
    "color_picker_color_slateblue": "Azul pizarra",
    "color_picker_color_slategray": "Gris pizarra",
    "color_picker_color_snow": "Nieve",
    "color_picker_color_springgreen": "Verde primavera",
    "color_picker_color_steelblue": "Azul acero",
    "color_picker_color_tan": "Bronceado",
    "color_picker_color_thistle": "Cardo",
    "color_picker_color_tomato": "Tomate",
    "color_picker_color_turquoise": "Turquesa",
    "color_picker_color_violet": "Violeta",
    "color_picker_color_wheat": "Trigo",
    "color_picker_color_whitesmoke": "Humo blanco",
    "color_picker_color_yellowgreen": "Verde amarillo",
}



# English lexemes settings_dialog.py
lexemes = {
    "tab_openai_api_config": "OpenAI API",

    "module_openai_api_label": "OpenAI API",
    "module_openai_api_url_label": "API URL",
    "module_openai_api_url_input_placeholder_text": "API URL",
    "module_openai_api_url_input_accessible_description":
        "The OpenAI API URL is the address to the API endpoint, which may vary depending on the service and version.\n"
        "The AI Assistant uses the one for conversational chat functionality or text completions.\n"
        "Refer to the official OpenAI API documentation to obtain the current URL.",
    "module_openai_api_key_label": "API Key",
    "module_openai_api_key_input_placeholder_text": "API Key",
    "module_openai_api_key_input_accessible_description":
        "The OpenAI API Key is a secret token used for authenticating requests to the API endpoint.",
    "module_openai_api_supported_models_label": "Supported Models",
    "module_openai_api_model_names_combo_placeholder_text": "Choose a Model",
    "module_openai_api_model_names_combo_accessible_description":
        "Select from supported models for chat conversations.",

    "module_openai_api_base_system_prompt_label": "System Prompt",
    "module_openai_api_base_system_prompt_edit_placeholder_text": "Base system prompt that precedes each request",
    "module_openai_api_base_system_prompt_edit_accessible_description":
        "A base system prompt that precedes each request.\n"
        "Usually, it is plain text with instructions or role characteristics.",

    "module_openai_api_base_response_temperature_label": "Temperature: {temperature}",
    "module_openai_api_base_response_temperature_input_accessible_description":
        "Adjusts the randomness of the model's responses. Higher values produce more varied outputs, "
        "while lower values make responses more predictable.",

    "module_openai_api_base_response_max_tokens_label": "Maximum Response Tokens",
    "module_openai_api_base_response_max_tokens_input_accessible_description":
        "Maximum number of tokens to receive in response, such as words and punctuation, "
        "controlling the length of the output.",

    "module_openai_api_config_prompt_history_size_label": "Prompt History Size",
    "module_openai_api_config_prompt_history_size_input_accessible_description":
        "Controls the number of entries in the prompt history that the system retains for reference.\n"
        "A zero value allows for unlimited entries."
}



# Korean lexemes common.py
lexemes = {
    "app_title": "Notolog 에디터",
    "app_title_with_sub": "{app_title} - {sub_title}",

    "tree_filter_accessible_desc": "파일 필터 필드",

    "menu_action_copy_file_path": "파일 경로 복사",
    "menu_action_rename": "이름 바꾸기",
    "menu_action_delete": "삭제",
    "menu_action_delete_completely": "완전히 삭제",
    "menu_action_restore": "복원",
    "menu_action_create_new_dir": "새 디렉터리 만들기",

    "dialog_file_rename_title": "파일 이름 바꾸기",
    "dialog_file_rename_field_label": "새 파일 이름 입력",
    "dialog_file_rename_button_ok": "이름 바꾸기",
    "dialog_file_rename_warning_exists": "동일한 이름의 파일이 이미 존재합니다",

    "dialog_file_delete_title": "파일 삭제",
    "dialog_file_delete_text": "파일 \"{file_name}\"을(를) 삭제하시겠습니까?",
    "dialog_file_delete_completely_title": "파일 완전 삭제",
    "dialog_file_delete_completely_text": "파일 \"{file_name}\"을(를) 완전히 삭제하시겠습니까?",
    "dialog_file_delete_error": "파일을 삭제할 수 없습니다. 오류가 발생했습니다",
    "dialog_file_delete_error_not_found": "파일을 찾을 수 없습니다",

    "dialog_file_restore_title": "파일 복원",
    "dialog_file_restore_text": "파일 \"{file_name}\"을(를) 복원하시겠습니까?",
    "dialog_file_restore_error": "파일을 복원할 수 없습니다. 오류가 발생했습니다",
    "dialog_file_restore_warning_exists": "{file_name} 이름의 파일이 이미 존재합니다",

    "dialog_create_new_dir_title": "새 디렉터리 만들기",
    "dialog_create_new_dir_label": "새 디렉터리 이름",
    "dialog_create_new_dir_input_placeholder_text": "디렉터리 이름 입력",
    "dialog_create_new_dir_button_ok": "만들기",
    "dialog_create_new_dir_button_cancel": "취소",
    "dialog_create_new_dir_warning_empty_title": "새 디렉터리 이름 오류",
    "dialog_create_new_dir_warning_empty_text": "디렉터리 이름은 비워둘 수 없습니다",
    "dialog_create_new_dir_warning_too_long_title": "새 디렉터리 이름 오류",
    "dialog_create_new_dir_warning_too_long_text": "디렉터리 이름이 너무 깁니다. 최대 "
                                                   "{symbols} 자까지 허용됩니다!",
    "dialog_create_new_dir_error_existed": "디렉터리가 이미 존재합니다",
    "dialog_create_new_dir_error": "디렉터리를 만들 수 없습니다. 대상 디렉터리 "
                                   "{base_dir} 이 쓰기 가능한지 확인하십시오",

    "dialog_message_box_title": "메시지",
    "dialog_message_box_button_ok": "닫기",

    "action_new_file_first_line_template_text": "새 문서",
    "action_open_file_dialog_caption": "파일 열기",
    "action_save_as_file_dialog_caption": "파일 저장",

    "dialog_save_empty_file_title": "빈 파일 저장",
    "dialog_save_empty_file_text": "내용이 없는 파일을 저장하시겠습니까?",

    "dialog_encrypt_file_title": "파일 암호화",
    "dialog_encrypt_file_text": "파일 \"{file_name}\"을(를) 암호화하시겠습니까?",
    "encrypt_file_warning_file_is_already_encrypted": "파일이 이미 암호화되었습니다!",
    "dialog_encrypt_file_rewrite_existing_title": "기존 파일 덮어쓰기",
    "dialog_encrypt_file_rewrite_existing_text": "기존 파일 \"{file_path}\"을(를) 덮어쓰시겠습니까?",

    "dialog_decrypt_file_title": "파일 복호화",
    "dialog_decrypt_file_text": "파일 \"{file_name}\"을(를) 복호화하시겠습니까?",
    "decrypt_file_warning_file_is_not_encrypted": "파일이 암호화되지 않았습니다!",
    "dialog_decrypt_file_rewrite_existing_title": "기존 파일 덮어쓰기",
    "dialog_decrypt_file_rewrite_existing_text": "기존 파일 \"{file_path}\"을(를) 덮어쓰시겠습니까?",

    "dialog_encrypt_new_password_title": "새 비밀번호",
    "dialog_encrypt_new_password_label": "비밀번호:",
    "dialog_encrypt_new_password_input_placeholder_text": "새 비밀번호 입력",
    "dialog_encrypt_new_password_hint_label": "힌트:",
    "dialog_encrypt_new_password_hint_label_description": "힌트는 암호화되지 않으며 파일에서 읽을 수 있습니다!"
                                                          "\n생일 등 쉽게 추측할 수 있는 힌트를 사용하지 마세요."
                                                          "\n관련 없는 참조를 사용해 보세요.",
    "dialog_encrypt_new_password_hint_input_placeholder_text": "힌트 입력(선택 사항)",
    "dialog_encrypt_new_password_button_ok": "확인",
    "dialog_encrypt_new_password_button_cancel": "취소",
    "dialog_encrypt_new_password_warning_empty_title": "경고",
    "dialog_encrypt_new_password_warning_empty_text": "비밀번호 필드는 비워 둘 수 없습니다!",
    "dialog_encrypt_new_password_warning_too_long_title": "경고",
    "dialog_encrypt_new_password_warning_too_long_text": "힌트 필드가 너무 깁니다, 최대 {symbols}개의 기호!",

    "dialog_encrypt_password_title": "비밀번호 입력",
    "dialog_encrypt_password_label": "비밀번호:",
    "dialog_encrypt_password_input_placeholder_text": "비밀번호 입력",
    "dialog_encrypt_password_hint_label": "힌트:",
    "dialog_encrypt_password_button_ok": "확인",
    "dialog_encrypt_password_button_cancel": "취소",

    "dialog_encrypt_password_reset_title": "암호화 비밀번호 재설정",
    "dialog_encrypt_password_reset_text": "현재 암호화 비밀번호를 재설정하시겠습니까?",
    "dialog_encrypt_password_reset_button_cancel": "취소",
    "dialog_encrypt_password_reset_button_yes": "예",

    "dialog_open_link_title": "링크",
    "dialog_open_link_text": "링크 \"{url}\"을(를) 브라우저에서 열겠습니까?",

    "dialog_reset_settings_title": "설정을 재설정하시겠습니까?",
    "dialog_reset_settings_text": "설정에 저장된 모든 데이터가 삭제되며, 변경 사항을 적용하기 위해 앱이 재시작됩니다.",

    "dialog_exit_unsaved_title": "종료 확인",
    "dialog_exit_unsaved_text": "열린 파일 '{file_name}'을(를) 저장할 수 없습니다. 종료하시겠습니까?",

    "message_app_config_file_access": "{file_path}에서 앱 구성 파일에 접근할 때 권한이 거부되었습니다. "
                                      "제대로된 작동을 보장하기 위해 올바른 권한을 설정하세요.",

    "field_dir_path_dialog_caption": "디렉토리 선택",
    "field_file_path_dialog_caption": "파일 선택",

    "load_file_encryption_password_mismatch": "암호화 비밀번호가 일치하지 않습니다!",
    "load_file_encryption_password_incorrect": "암호화 비밀번호가 정확하지 않습니다!",
    "load_file_none_content_error": "파일을 불러올 수 없습니다.",

    "action_new_file_error_occurred": "파일을 생성할 수 없습니다; 오류가 발생했습니다。\n파일 시스템 권한을 확인하십시오。",
    "save_active_file_error_occurred": "파일을 저장할 수 없습니다; 오류가 발생했습니다。",

    "expandable_block_default_title": "추가 정보...",
    "expandable_block_open_close_tags_mismatch_warning": "<details> 블록 열기/닫기 태그 불일치",

    "dialog_color_picker_color_copied_to_the_clipboard": "서식이 지정된 텍스트가 클립보드에 복사되었습니다",

    "popup_about_title": "애플리케이션 정보",
    "popup_about_app_name_description": "파이썬 마크다운 에디터",

    "popup_about_version": "버전",
    "popup_about_license": "라이선스",
    "popup_about_website": "웹사이트",
    "popup_about_repository": "GitHub",
    "popup_about_pypi": "PyPi",
    "popup_about_date": "날짜",

    "update_helper_new_version_is_available": "앱의 새 버전 {latest_version}을(를) 사용할 수 있습니다.",
    "update_helper_latest_version_installed": "앱의 최신 버전이 설치되었습니다",

    "network_connection_error_empty": "응답 정보를 가져올 수 없습니다",
    "network_connection_error_connection_or_dns":
        "호스트를 찾을 수 없습니다. 인터넷 연결 또는 DNS에 문제가 있을 수 있습니다.",
    "network_connection_error_connection_refused":
        "연결이 거부되었습니다. 서버가 다운되었거나 네트워크에 문제가 있을 수 있습니다.",
    "network_connection_error_connection_timed_out": "연결 시간 초과. 네트워크에 문제가 있을 수 있습니다.",
    "network_connection_error_connection_404_error":
        "404 연결 오류. 요청한 페이지 또는 리소스를 찾을 수 없습니다.",
    "network_connection_error_generic_with_status_code": "상태 코드 {status_code}으로 요청 실패",
}



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Enhances QLineEdit functionality by adding an icon that triggers a QFileDialog action for file selection.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QStyle, QLineEdit, QPushButton, QFileDialog
from PySide6.QtGui import QColor

from . import Lexemes
from . import ThemeHelper

import logging


class FilePathLineEdit(QLineEdit):

    button: QPushButton = None

    def __init__(self, parent=None, settings=None, default_file=None, ext_filter=None):
        super(FilePathLineEdit, self).__init__(parent)

        self.parent = parent

        if self.parent and hasattr(self.parent, 'font'):
            # Apply font from the main window to the element
            self.setFont(self.parent.font())

        # The class may be in use within settings itself
        self.settings = settings if settings else self.parent.settings  # type: ignore

        self.logger = logging.getLogger('file_path_line_edit')

        # Load lexemes for selected language and scope
        self.lexemes = Lexemes(self.settings.app_language, default_scope='common')

        self.theme_helper = ThemeHelper()

        self.default_file = default_file
        # "All Files (*.*)", "Text Files (*.txt)", etc.
        # Or a combined filter set like:
        # "Text Files (*.txt);;Image Files (*.png *.jpg);;All Files (*.*)"
        self.ext_filter = ext_filter

        self.init_ui()

    def init_ui(self):

        self.button = QPushButton(self)

        # Load the dialog icon
        self.load_icon()

        self.button.clicked.connect(self.select_file_dialog)
        self.button.setCursor(Qt.CursorShape.ArrowCursor)
        self.button.setStyleSheet("QPushButton { border: none; padding: 0; }")
        self.button.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        # Set the default file path if provided
        if self.default_file:
            self.setText(self.default_file)

        # Ensure the text doesn't overlap the button
        self.setTextMargins(0, 0, 30, 0)

    def load_icon(self):
        # Use either a standard or a custom icon for file selection.
        # file_icon = self.style().standardIcon(QStyle.StandardPixmap.SP_FileIcon)  # Set an icon using standard icons
        icon_color = self.theme_helper.get_color('settings_dialog_hint_icon_color')
        file_icon = self.theme_helper.get_icon(theme_icon='files-alt.svg', color=QColor(icon_color))
        self.button.setIcon(file_icon)

    def resizeEvent(self, event):
        button_size = self.button.sizeHint()
        frame_width = self.style().pixelMetric(QStyle.PixelMetric.PM_DefaultFrameWidth)
        self.button.move(self.rect().right() - frame_width - button_size.width(),
                         (self.rect().bottom() - button_size.height() + 1) // 2)
        super(FilePathLineEdit, self).resizeEvent(event)

    def select_file_dialog(self):
        file_path, _ = QFileDialog.getOpenFileName(self, caption=self.lexemes.get('field_file_path_dialog_caption'),
                                                   dir=self.text(), filter=self.ext_filter)
        if file_path:
            self.setText(file_path)



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: File tree context menu that displays context actions for files to the user.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtGui import QAction, QColor
from PySide6.QtWidgets import QMenu

from . import Settings
from . import Lexemes
from . import ThemeHelper
from . import ClipboardHelper

from ..ui.create_new_dir_dialog import CreateNewDirDialog

import os
import logging


class FileTreeContextMenu(QMenu):
    def __init__(self, file_path: str = None, parent=None):
        super().__init__(parent)

        self.parent = parent
        # Target file path
        self.file_path = file_path

        # Apply font from the parent instance to the menu
        self.setFont(self.parent.font())

        self.settings = Settings(parent=self)

        self.logger = logging.getLogger('file_tree_context_menu')

        # Load lexemes for selected language and scope
        self.lexemes = Lexemes(self.settings.app_language, default_scope='common')

        # Theme helper
        self.theme_helper = ThemeHelper()

        self.init_ui()

    def init_ui(self):
        # menu = QMenu(self)
        # Font is already set in constructor
        # menu.setFont(self.font())

        current_dir_path = self.parent.get_tree_active_dir()

        if self.file_path:
            if os.path.isfile(self.file_path):
                self.file_menu()
            elif os.path.isdir(self.file_path):
                # Create a sub-directory
                current_dir_path = self.file_path

        self.tree_menu(dir_path=current_dir_path)

    def file_menu(self):
        # Copy file path context action
        copy_file_path_icon = self.theme_helper.get_icon(
            theme_icon='signpost-split.svg', system_icon='edit-copy',
            color=QColor(self.theme_helper.get_color('main_tree_context_menu_copy_file_path')))
        self.addAction(copy_file_path_icon, self.lexemes.get('menu_action_copy_file_path'),
                       lambda: self.copy_file_path_dialog(self.file_path))

        # Rename file context action
        rename_icon = self.theme_helper.get_icon(
            theme_icon='cursor-text.svg', system_icon='document-properties',
            color=QColor(self.theme_helper.get_color('main_tree_context_menu_rename')))
        self.addAction(rename_icon, self.lexemes.get('menu_action_rename'),
                       lambda: self.parent.rename_file_dialog(self.file_path))

        if self.parent.is_file_safely_deleted(self.file_path):
            # Restore deleted file context action
            restore_icon = self.theme_helper.get_icon(
                theme_icon='box-arrow-up.svg', system_icon='edit-undo',
                color=QColor(self.theme_helper.get_color('main_tree_context_menu_restore')))
            self.addAction(restore_icon, self.lexemes.get('menu_action_restore'),
                           lambda: self.parent.restore_file_dialog(self.file_path))
            # Delete completely file context action
            delete_icon = self.theme_helper.get_icon(
                theme_icon='x-square-fill.svg', system_icon='edit-delete',
                color=QColor(self.theme_helper.get_color('main_tree_context_menu_delete_completely')))
            self.addAction(delete_icon, self.lexemes.get('menu_action_delete_completely'),
                           lambda: self.parent.delete_completely_file_dialog(self.file_path))
        else:
            # Delete file context action
            delete_icon = self.theme_helper.get_icon(
                theme_icon='x-square.svg', system_icon='edit-delete',
                color=QColor(self.theme_helper.get_color('main_tree_context_menu_delete')))
            self.addAction(delete_icon, self.lexemes.get('menu_action_delete'),
                           lambda: self.parent.delete_file_dialog(self.file_path))

    def tree_menu(self, dir_path: str):
        # Create new dir context action
        create_new_dir_icon = self.theme_helper.get_icon(
            theme_icon='folder-plus.svg', system_icon='folder-new',
            color=QColor(self.theme_helper.get_color('main_tree_context_menu_create_new_dir')))
        create_new_dir_action = QAction(self.lexemes.get('menu_action_create_new_dir'), self)
        create_new_dir_action.setIcon(create_new_dir_icon)
        create_new_dir_action.triggered.connect(lambda: self.create_new_dir_dialog(base_dir=dir_path))
        # Set as not accessible if no write access
        if not os.access(dir_path, os.W_OK):
            create_new_dir_action.setEnabled(False)
        # Add the action to the toolbar
        self.addAction(create_new_dir_action)

    def copy_file_path_dialog(self, file_path):
        # Copy text to the clipboard
        ClipboardHelper.set_text(file_path)

    def create_new_dir_dialog(self, base_dir):
        # Show create a new folder dialog
        dialog = CreateNewDirDialog(base_dir, parent=self.parent)
        dialog.exec()



# German lexemes ai_assistant.py
lexemes = {
    # KI-Assistent
    "dialog_title": "Anfrage an KI",
    "dialog_prompt_input_placeholder_text": "Anfrage",
    "dialog_prompt_input_accessible_description": "Anfrage eingeben, um dem KI-Assistenten Kontext zu geben",
    "dialog_response_output_placeholder_text": "Antwortdaten",
    "dialog_response_output_accessible_description": "Antwortdaten erscheinen in diesem Feld",
    "dialog_response_output_notice_empty_text": "Geben Sie einen Text ein, um eine Antwort von etwa...",

    "dialog_usage_module_label": "MODUL",
    "dialog_usage_model_label": "MODELL",
    "dialog_usage_tokens_label": "TOKENS",
    "dialog_usage_tokens_prompt": "prompt: {tokens}",
    "dialog_usage_tokens_answer": "Antwort: {tokens}",
    "dialog_usage_tokens_total": "gesamt: {tokens}",

    "dialog_button_send_request": "Anfrage senden",
    "dialog_button_stop_request": "Stop",
    "dialog_button_save_history": "Dialogverlauf speichern",

    "dialog_message_copy_tooltip": "Text kopieren",
    "dialog_message_copied_tooltip": "Kopiert",

    "dialog_error_loading_model": "Fehler beim Laden des benutzerdefinierten Modells",
}



# Korean lexemes common.py
lexemes = {
    # Common
    "module_ondevice_llm_name": "기기 내 LLM",

    "module_ondevice_llm_model_exception": "모델을 찾을 수 없음 오류: {error_msg}",
    "module_ondevice_llm_task_exception": "모델 생성기를 초기화할 수 없습니다: {error_msg}",
}



# Finnish lexemes toolbar.py
lexemes = {
    "actions_label_new_document": "Uusi asiakirja",
    "actions_accessible_name_new_document": "Uusi asiakirja",
    "actions_label_open": "Avaa",
    "actions_accessible_name_open": "Avaa asiakirja",
    "actions_label_save": "Tallenna",
    "actions_accessible_name_save": "Tallenna asiakirja",
    "actions_label_save_as": "Tallenna nimellä",
    "actions_accessible_name_save_as": "Tallenna asiakirjan kopio",
    "actions_label_sync": "Synkronoi",
    "actions_accessible_name_sync": "Synkronoi asiakirja ulkoiseen tallennustilaan",
    "actions_label_edit": "Muokkaa",
    "actions_accessible_name_edit": "Muokkaa tiedostoa",
    "actions_label_source": "Lähdekoodi",
    "actions_accessible_name_source": "Lähdekoodi",
    "actions_label_encrypt": "Salaa",
    "actions_accessible_name_encrypt": "Salaa tiedosto",
    "actions_label_decrypt": "Pura salaus",
    "actions_accessible_name_decrypt": "Pura tiedoston salaus",
    "actions_label_bold": "Lihavointi",
    "actions_accessible_name_bold": "Lihavoi teksti",
    "actions_label_italic": "Kursivointi",
    "actions_accessible_name_italic": "Kursivoi teksti",
    "actions_label_underline": "Alleviivaus",
    "actions_accessible_name_underline": "Alleviivaa teksti",
    "actions_label_strikethrough": "Yliviivaus",
    "actions_accessible_name_strikethrough": "Yliviivaa teksti",
    "actions_label_blockquote": "Lainaus",
    "actions_accessible_name_blockquote": "Lainaa teksti",
    "actions_label_color_picker": "Värinvalitsin",
    "actions_accessible_name_color_picker": "Valitse tekstiväri",
    "actions_label_ai_assistant": "Tekoälyavustaja",
    "actions_accessible_name_ai_assistant": "Käynnistä tekoälyavustaja",
    "actions_label_about": "Tietoja",
    "actions_accessible_name_about": "Tietoja sovelluksesta",
    "actions_label_exit": "Poistu",
    "actions_accessible_name_exit": "Poistu sovelluksesta",
    "actions_label_settings": "Asetukset",
    "actions_accessible_name_settings": "Sovelluksen asetukset",

    "search_input_label": "Haku",
    "search_input_placeholder_text": "Hae tekstiä dokumentista",
    "search_input_accessible_description": "Hakukenttä",
    "search_buttons_label_clear": "Tyhjennä haku",
    "search_buttons_accessible_name_clear": "Tyhjennä hakukenttä",
    "search_buttons_label_prev": "Edellinen haku",
    "search_buttons_accessible_name_prev": "Edellinen hakutulos",
    "search_buttons_label_next": "Seuraava haku",
    "search_buttons_accessible_name_next": "Seuraava hakutulos",
    "search_case_sensitive_accessible_name": "Kirjainkoko merkitsee",
    "search_case_sensitive_tooltip": "Haku kirjainkoolla",
}



# Georgian lexemes common.py
lexemes = {
    "app_title": "Notolog რედაქტორი",
    "app_title_with_sub": "{app_title} - {sub_title}",

    "tree_filter_accessible_desc": "ფაილის ფილტრის ველი",

    "menu_action_copy_file_path": "ფაილის გზის კოპირება",
    "menu_action_rename": "გადარქმევა",
    "menu_action_delete": "წაშლა",
    "menu_action_delete_completely": "სრული წაშლა",
    "menu_action_restore": "აღდგენა",
    "menu_action_create_new_dir": "ახალი დირექტორიის შექმნა",

    "dialog_file_rename_title": "ფაილის გადარქმევა",
    "dialog_file_rename_field_label": "შეიყვანეთ ახალი ფაილის სახელი",
    "dialog_file_rename_button_ok": "გადარქმევა",
    "dialog_file_rename_warning_exists": "ამ სახელით ფაილი უკვე არსებობს",

    "dialog_file_delete_title": "ფაილის წაშლა",
    "dialog_file_delete_text": "წაშლა ფაილი \"{file_name}\"?",
    "dialog_file_delete_completely_title": "ფაილის სრული წაშლა",
    "dialog_file_delete_completely_text": "წაშალეთ ფაილი \"{file_name}\" სრულად?",
    "dialog_file_delete_error": "ფაილის წაშლა ვერ ხერხდება, წარმოიშვა შეცდომა",
    "dialog_file_delete_error_not_found": "ფაილი ვერ მოიძებნა",

    "dialog_file_restore_title": "ფაილის აღდგენა",
    "dialog_file_restore_text": "აღადგინოთ ფაილი \"{file_name}\"?",
    "dialog_file_restore_error": "ფაილის აღდგენა ვერ ხერხდება, წარმოიშვა შეცდომა",
    "dialog_file_restore_warning_exists": "ამ სახელით ფაილი უკვე არსებობს",

    "dialog_create_new_dir_title": "ახალი დირექტორიის შექმნა",
    "dialog_create_new_dir_label": "ახალი დირექტორიის სახელი",
    "dialog_create_new_dir_input_placeholder_text": "შეიყვანეთ დირექტორიის სახელი",
    "dialog_create_new_dir_button_ok": "შექმნა",
    "dialog_create_new_dir_button_cancel": "გაუქმება",
    "dialog_create_new_dir_warning_empty_title": "ახალი დირექტორიის სახელის შეცდომა",
    "dialog_create_new_dir_warning_empty_text": "დირექტორიის სახელი არ შეიძლება იყოს ცარიელი",
    "dialog_create_new_dir_warning_too_long_title": "ახალი დირექტორიის სახელის შეცდომა",
    "dialog_create_new_dir_warning_too_long_text": "დირექტორიის სახელი ძალიან გრძელია; მაქსიმუმ "
                                                   "{symbols} სიმბოლოებია ნებადართული!",
    "dialog_create_new_dir_error_existed": "დირექტორია უკვე არსებობს",
    "dialog_create_new_dir_error": "ვერ შევქმენით დირექტორია. დარწმუნდით, რომ სამიზნე დირექტორია "
                                   "{base_dir} ჩაწერის შესაძლებლობით არის",

    "dialog_message_box_title": "შეტყობინება",
    "dialog_message_box_button_ok": "დახურვა",

    "action_new_file_first_line_template_text": "ახალი დოკუმენტი",
    "action_open_file_dialog_caption": "ფაილის გახსნა",
    "action_save_as_file_dialog_caption": "ფაილის შენახვა",

    "dialog_save_empty_file_title": "ცარიელი ფაილის შენახვა",
    "dialog_save_empty_file_text": "დაშვებულია ფაილის შენახვა ცარიელი შინაარსით?",

    "dialog_encrypt_file_title": "ფაილის დაშიფვრა",
    "dialog_encrypt_file_text": "დაშიფრეთ ფაილი \"{file_name}\"?",
    "encrypt_file_warning_file_is_already_encrypted": "ფაილი უკვე დაშიფრულია!",
    "dialog_encrypt_file_rewrite_existing_title": "არსებული ფაილის გადაწერა",
    "dialog_encrypt_file_rewrite_existing_text": "გადაწერეთ არსებული ფაილი \"{file_path}\"?",

    "dialog_decrypt_file_title": "ფაილის დეშიფვრა",
    "dialog_decrypt_file_text": "დეშიფრეთ ფაილი \"{file_name}\"?",
    "decrypt_file_warning_file_is_not_encrypted": "ფაილი არ არის დაშიფრული!",
    "dialog_decrypt_file_rewrite_existing_title": "არსებული ფაილის გადაწერა",
    "dialog_decrypt_file_rewrite_existing_text": "გადაწერეთ არსებული ფაილი \"{file_path}\"?",

    "dialog_encrypt_new_password_title": "ახალი პაროლი",
    "dialog_encrypt_new_password_label": "პაროლი:",
    "dialog_encrypt_new_password_input_placeholder_text": "შეიყვანეთ ახალი პაროლი",
    "dialog_encrypt_new_password_hint_label": "მინიშნება:",
    "dialog_encrypt_new_password_hint_label_description":
        "მინიშნება არ არის დაშიფრული და ფაილიდან იკითხება!"
        "\nარ გამოიყენოთ მინიშნები, რომლებიც ადვილად შეიმჩნევა, როგორიცაა"
        "\nდაბადების თარიღი."
        "\nსცადეთ გამოიყენოთ რეფერენცია, რომელიც ადვილად არ არის თქვენთან გაიგიანებელი.",
    "dialog_encrypt_new_password_hint_input_placeholder_text": "შეიყვანეთ მინიშნება (არასავალდებულო)",
    "dialog_encrypt_new_password_button_ok": "კარგი",
    "dialog_encrypt_new_password_button_cancel": "გაუქმება",
    "dialog_encrypt_new_password_warning_empty_title": "გაფრთხილება",
    "dialog_encrypt_new_password_warning_empty_text": "პაროლის ველი ცარიელი ვერ დარჩება!",
    "dialog_encrypt_new_password_warning_too_long_title": "გაფრთხილება",
    "dialog_encrypt_new_password_warning_too_long_text": "მინიშნების ველი ძალიან გრძელია, მაქსიმუმი {symbols} სიმბოლო!",

    "dialog_encrypt_password_title": "პაროლის შეტანა",
    "dialog_encrypt_password_label": "პაროლი:",
    "dialog_encrypt_password_input_placeholder_text": "შეიყვანეთ პაროლი",
    "dialog_encrypt_password_hint_label": "მინიშნება:",
    "dialog_encrypt_password_button_ok": "კარგი",
    "dialog_encrypt_password_button_cancel": "გაუქმება",

    "dialog_encrypt_password_reset_title": "დაშიფრვის პაროლის განულება",
    "dialog_encrypt_password_reset_text": "დარწმუნებული ხართ, რომ გსურთ მიმდინარე დაშიფრვის პაროლის განულება?",
    "dialog_encrypt_password_reset_button_cancel": "გაუქმება",
    "dialog_encrypt_password_reset_button_yes": "დიახ",

    "dialog_open_link_title": "ბმული",
    "dialog_open_link_text": "გახსნა ბმული \"{url}\" ბრაუზერში?",

    "dialog_reset_settings_title": "პარამეტრების აღდგენა?",
    "dialog_reset_settings_text":
        "ყველა შენახული მონაცემი პარამეტრებში წაიშლება, და აპლიკაცია გადაიტვირთება ცვლილებების გამო.",

    "dialog_exit_unsaved_title": "გასვლის დადასტურება",
    "dialog_exit_unsaved_text": "გახსნილი ფაილი '{file_name}' ვერ ინახება. გავაგრძელოთ გასვლა?",

    "message_app_config_file_access": "დაშვების უფლებები უარყოფილია {file_path} აპლიკაციის კონფიგურაციული ფაილის წვდომისას. "
                                      "დააყენეთ სწორი უფლებები სწორი ოპერაციის გარანტირებლად.",

    "field_dir_path_dialog_caption": "აირჩიეთ დირექტორია",
    "field_file_path_dialog_caption": "აირჩიეთ ფაილი",

    "load_file_encryption_password_mismatch": "დაშიფრვის პაროლის უთანასწორობა!",
    "load_file_encryption_password_incorrect": "დაშიფრვის პაროლი არასწორია!",
    "load_file_none_content_error": "ფაილი ვერ იტვირთება.",

    "action_new_file_error_occurred": "ფაილის შექმნა ვერ ხერხდება; მოხდა შეცდომა.\nგადამოწმეთ ფაილის სისტემის ნებართვები.",
    "save_active_file_error_occurred": "ფაილის შენახვა ვერ ხერხდება; მოხდა შეცდომა.",

    "expandable_block_default_title": "მეტი ინფორმაცია...",
    "expandable_block_open_close_tags_mismatch_warning": "<details> ბლოკის გახსნის/დახურვის თეგების არასწორი მოთხოვნები",

    "dialog_color_picker_color_copied_to_the_clipboard": "ფორმატირებული ტექსტი ბუფერში ასლი გადაიტანეს",

    "popup_about_title": "აპლიკაციის ინფორმაცია",
    "popup_about_app_name_description": "Python-ის Markdown რედაქტორი",

    "popup_about_version": "ვერსია",
    "popup_about_license": "ლიცენზია",
    "popup_about_website": "ვებგვერდი",
    "popup_about_repository": "GitHub",
    "popup_about_pypi": "PyPi",
    "popup_about_date": "თარიღი",

    "update_helper_new_version_is_available": "აპლიკაციის ახალი ვერსია {latest_version} მისაწვდომია",
    "update_helper_latest_version_installed": "აპლიკაციის უახლესი ვერსია დაყენებულია",

    "network_connection_error_empty": "პასუხის ინფორმაციის მიღება ვერ ხერხდება",
    "network_connection_error_connection_or_dns":
        "ჰოსტი ვერ მოიძებნა. შესაძლოა ინტერნეტის კავშირში ან DNS-ში იყოს პრობლემა.",
    "network_connection_error_connection_refused":
        "კავშირი უარყოფილია. სერვერი შესაძლოა გამორთული იყოს ან ქსელში იყოს პრობლემები.",
    "network_connection_error_connection_timed_out": "კავშირი დავადებით დამთავრდა. შესაძლოა ქსელში იყოს პრობლემები.",
    "network_connection_error_connection_404_error":
        "404 კავშირის შეცდომა. მოთხოვნილი გვერდი ან რესურსი ვერ მოიძებნა.",
    "network_connection_error_generic_with_status_code": "მოთხოვნა ვერ შესრულდა, სტატუსის კოდი: {status_code}",
}



# Finnish lexemes main_menu.py
lexemes = {
    "group_file_label": "Tiedosto",
    "group_edit_label": "Muokkaa",
    "group_tools_label": "Työkalut",
    "group_help_label": "Apua",

    "actions_file_label_new_document": "Uusi asiakirja",
    "actions_file_accessible_name_new_document": "Uusi asiakirja",
    "actions_file_label_open": "Avaa",
    "actions_file_accessible_name_open": "Avaa asiakirja",
    "actions_file_label_save": "Tallenna",
    "actions_file_accessible_name_save": "Tallenna asiakirja",
    "actions_file_label_save_as": "Tallenna nimellä",
    "actions_file_accessible_name_save_as": "Tallenna asiakirjan kopio",
    "actions_file_label_settings": "Asetukset",
    "actions_file_accessible_name_settings": "Sovelluksen asetukset",
    "actions_file_label_reset_settings": "Nollaa asetukset",
    "actions_file_accessible_name_reset_settings": "Nollaa asetukset",
    "actions_file_label_exit": "Poistu",
    "actions_file_accessible_name_exit": "Poistu sovelluksesta",

    "actions_edit_edit_mode": "Muokkaa",
    "actions_edit_accessible_name_edit_mode": "Muokkaa tiedostoa",
    "actions_edit_source_mode": "Lähdekoodi",
    "actions_edit_accessible_name_source_mode": "Lähdekoodi",
    "actions_edit_bold": "Lihavoitu",
    "actions_edit_accessible_name_bold": "Lihavoi teksti",
    "actions_edit_italic": "Kursivoitu",
    "actions_edit_accessible_name_italic": "Kursivoi teksti",
    "actions_edit_underline": "Alleviivaus",
    "actions_edit_accessible_name_underline": "Alleviivaa teksti",
    "actions_edit_strikethrough": "Yliviivaus",
    "actions_edit_accessible_name_strikethrough": "Yliviivaa teksti",
    "actions_edit_blockquote": "Lainaus",
    "actions_edit_accessible_name_blockquote": "Lainaa teksti",

    "actions_tools_ai_assistant": "Tekoälyavustaja",
    "actions_tools_accessible_name_ai_assistant": "Käynnistä tekoälyavustaja",
    "actions_tools_color_picker": "Värinvalitsin",
    "actions_tools_accessible_name_color_picker": "Valitse tekstiväri",

    "actions_help_label_md_syntax": "Markdown-syntaksi",
    "actions_help_accessible_name_md_syntax": "Markdown-syntaksi",
    "actions_help_label_check_for_updates": "Tarkista päivitykset",
    "actions_help_accessible_name_check_for_updates": "Tarkista päivitykset",
    "actions_help_label_bug_report":  "Vikailmoitus",
    "actions_help_accessible_name_bug_report": "Lähetä vikailmoitus",
    "actions_help_label_about": "Tietoja",
    "actions_help_accessible_name_about": "Tietoja",
}



# Turkish lexemes settings_dialog.py
lexemes = {
    "window_title": "Ayarlar",

    "button_close": "Kapat",

    "tab_general": "Genel",
    "tab_editor_config": "Editör",
    "tab_viewer_config": "Görüntüleyici",
    "tab_ai_config": "AI Ayarları",

    "general_app_config_label": "Uygulama ayarları",
    "general_app_language_label": "Dil",
    "general_app_language_combo_placeholder_text": "Bir dil seçin",
    "general_app_language_combo_accessible_description": "Uygulamanın arayüz dili",
    "general_app_theme_label": "Tema",
    "general_app_theme_combo_placeholder_text": "Bir tema seçin",
    "general_app_theme_combo_accessible_description": "Uygulamanın arayüz teması",
    "general_app_main_menu_label": "Ana Menü",
    "general_app_main_menu_checkbox": "Ana Menüyü Göster",
    "general_app_main_menu_checkbox_accessible_description": "Uygulamanın ana açılır menüsünü göster",
    "general_app_font_size_label": "Yazı Boyutu: {size}pt",
    "general_app_font_size_slider_accessible_description": "Uygulamanın genel yazı boyutunu ayarlayın",

    "general_statusbar_label": "Durum Çubuğu",
    "general_statusbar_show_global_cursor_position_checkbox": "Genel İmleç Pozisyonunu Göster",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Durum çubuğunda genel imleç pozisyonunu göster",

    "editor_config_label": "Editör Yapılandırması",
    "editor_config_show_line_numbers_checkbox": "Satır Numaralarını Göster",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Editörde satır numaralarını göster",

    "viewer_config_label": "Görüntüleyici Yapılandırması",
    "viewer_config_process_emojis_checkbox": "Metin Emojilerini Grafiklere Dönüştür",
    "viewer_config_process_emojis_checkbox_accessible_description": "Metin emojilerini grafiksel temsillere dönüştür",
    "viewer_config_highlight_todos_checkbox": "TODO'ları Vurgula",
    "viewer_config_highlight_todos_checkbox_accessible_description": "Metindeki TODO etiketlerini vurgula",
    "viewer_config_open_link_confirmation_checkbox": "Bağlantıları Açmadan Önce Onay İste",
    "viewer_config_open_link_confirmation_checkbox_accessible_description": "Bağlantıları açmadan önce onay isteyin",
    "viewer_config_save_resources_checkbox": "Dış Resimleri Diskte Otomatik Kaydet",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Dış resimlerin kopyalarını çevrimdışı erişim için diske otomatik olarak kaydet",

    "ai_config_inference_module_label": "Çıkarım Modülü",
    "ai_config_inference_module_names_combo_label": "Aktif Çıkarım Modülü",
    "ai_config_inference_module_names_combo_placeholder_text": "Modül Seçin",
    "ai_config_inference_module_names_combo_accessible_description":
        "AI Asistanı ile çalışacak mevcut AI çıkarım modüllerinden seçim yapın.\n"
        "Seçenekler, gerçek zamanlı işleme sahip büyük dil modelleri (LLM) veya API tabanlı işlevselliği içerir.",

    "ai_config_base_label": "Temel Parametreler",
    "ai_config_multi_turn_dialogue_checkbox": "Konuşma hafızasına sahip çok dönüşlü diyalog",
    "ai_config_multi_turn_dialogue_checkbox_accessible_description":
        "Konuşma hafızası için önceki istemi koruyan çok dönüşlü diyaloğu etkinleştirin.\n"
        "Devre dışı bırakıldığında, yalnızca yeni mesaj ve sistem istemi yanıtı etkiler.",
    "ai_config_convert_to_md_checkbox": "Sonucu Markdown'a Dönüştür",
    "ai_config_convert_to_md_checkbox_accessible_description":
        "Çıkış mesajını Markdown formatına dönüştürün.",
}



# Italian lexemes settings_dialog.py
lexemes = {
    "tab_module_llama_cpp_config": "Modulo llama.cpp",

    "module_llama_cpp_config_label": "Modulo llama.cpp",
    "module_llama_cpp_config_path_label": "Posizione del modello",
    "module_llama_cpp_config_path_input_placeholder_text": "Seleziona o inserisci il percorso del modello",
    "module_llama_cpp_config_path_input_accessible_description":
        "Un campo di inserimento con un selettore per specificare il percorso del modello locale. Supporta\n"
        "modelli nel formato GGUF, un formato di file binario ottimizzato per l'archiviazione di modelli\n"
        "utilizzati con GGML e esecutori basati su GGML.",
    "module_llama_cpp_config_path_input_filter_text": "File GGUF",

    "module_llama_cpp_config_context_window_label": "Dimensione della finestra di contesto",
    "module_llama_cpp_config_context_window_input_accessible_description":
        "Imposta il numero di token che il modello considera per generare risposte.\n"
        "Controlla quanta contestazione precedente viene utilizzata.",

    "module_llama_cpp_chat_formats_label": "Formati di chat",
    "module_llama_cpp_chat_formats_combo_placeholder_text": "Seleziona un formato di chat",
    "module_llama_cpp_chat_formats_combo_accessible_description":
        "Menu a tendina per selezionare il formato utilizzato per le conversazioni del modello.",

    "module_llama_cpp_config_system_prompt_label": "Prompt del sistema",
    "module_llama_cpp_config_system_prompt_edit_placeholder_text": "Inserisci il testo del prompt del sistema",
    "module_llama_cpp_config_system_prompt_edit_accessible_description":
        "Campo di testo per inserire i prompt del sistema che guidano le risposte del modello.",

    "module_llama_cpp_config_response_temperature_label": "Temperatura di risposta: {temperature}",
    "module_llama_cpp_config_response_temperature_input_accessible_description":
        "Regola la casualità delle risposte del modello. Valori più alti producono output più variati,\n"
        "mentre valori più bassi risultano in risposte più prevedibili.",

    "module_llama_cpp_config_response_max_tokens_label": "Token massimi per risposta",
    "module_llama_cpp_config_response_max_tokens_input_accessible_description":
        "Limita il numero di token nelle risposte del modello fino al limite effettivo della finestra di contesto.\n"
        "Un valore zero assume la capacità della finestra di contesto.",

    "module_llama_cpp_config_prompt_history_size_label": "Dimensione della cronologia dei prompt",
    "module_llama_cpp_config_prompt_history_size_input_accessible_description":
        "Controlla il numero di voci nella cronologia dei prompt mantenute dal sistema per riferimento.\n"
        "Un valore zero permette voci illimitate."
}



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Component of the 'Module llama.cpp' module.
- Functionality: Facilitates local inference for supported Large Language Models (LLMs)
                 in the GPT-Generated Unified Format (GGUF).

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import Qt, Signal
from PySide6.QtWidgets import QVBoxLayout, QWidget, QLabel, QSpinBox, QSlider, QSizePolicy, QPlainTextEdit, QScrollArea

import os
import logging

import asyncio
from asyncio import Task
from qasync import asyncSlot

from typing import Callable

from .model_helper import ModelHelper
from .prompt_manager import PromptManager

from ..base_ai_core import BaseAiCore

from .. import Settings
from .. import AppConfig
from .. import Lexemes

from ...ui.ai_assistant.ai_assistant import EnumMessageType, EnumMessageStyle
from ...ui.file_path_line_edit import FilePathLineEdit
from ...ui.horizontal_line_spacer import HorizontalLineSpacer
from ...ui.enum_combo_box import EnumComboBox
from ...ui.label_with_hint import LabelWithHint

from ...enums.llm_chat_formats import LlmChatFormats


class ModuleCore(BaseAiCore):
    update_signal = Signal(str, object, object, EnumMessageType, EnumMessageStyle)
    # prompt_tokens, completion_tokens, total_tokens
    update_usage_signal = Signal(str, object, object, object, object)

    # Default name, use lexeme for the translated version
    module_name = 'Module llama.cpp'

    # Functionality extended by this module
    extensions = ['ai_assistant', 'settings_dialog']

    init_callback: Callable = None
    finished_callback: Callable = None

    generator_task: Task = None

    prompt_manager: PromptManager = None

    def __init__(self):
        super(ModuleCore, self).__init__()

        self.settings = Settings(parent=self)
        self.settings.value_changed.connect(
            lambda v: self.settings_update_handler(v))

        self.logger = logging.getLogger('module_llama_cpp')

        # Load lexemes for the selected language and scope
        self.lexemes = Lexemes(self.settings.app_language,
                               default_scope='settings_dialog',
                               lexemes_dir=self.get_lexemes_path())

        # Explicitly extend settings to avoid circular recursion due to its global nature
        self.extend_settings_create_property(self.settings.create_property)
        # Register this module in the inference modules selection list
        inference_modules = self.settings.ai_config_inference_modules
        default_module = [key for key, value in inference_modules.items() if isinstance(value, tuple) and True in value]
        inference_modules.update({'llama_cpp': (self.lexemes.get('module_llama_cpp_name', scope='common'),
                                                True if not default_module else False)})
        self.settings.ai_config_inference_modules = inference_modules

        model_path = self.settings.module_llama_cpp_model_path
        context_window = self.settings.module_llama_cpp_context_window
        chat_format = self.settings.module_llama_cpp_chat_format
        search_options = self.get_search_options()
        # Cached helper instance for efficiency
        self.model_helper = ModelHelper(model_path=model_path, n_ctx=context_window, chat_format=chat_format,
                                        search_options=search_options)

        # Use for debugging asynchronous events if necessary:
        # asyncio.get_event_loop().set_debug(True)

        self.logger.debug(f'Module {__name__} loaded')

    def get_search_options(self):
        search_options = {}
        if hasattr(self.settings, 'module_llama_cpp_response_temperature'):
            search_options.update({'temperature': self.settings.module_llama_cpp_response_temperature})
        if hasattr(self.settings, 'module_llama_cpp_response_max_tokens'):
            search_options.update({'max_tokens': self.settings.module_llama_cpp_response_max_tokens})
        return search_options

    def init_prompt_manager(self, ai_dialog):
        """
        Prompt manager handles prompt management and appends messages to the prompt history.
        Note: It may not be initiated if called out of context, such as from settings.
        """
        self.prompt_manager = PromptManager(system_prompt=self.settings.module_llama_cpp_system_prompt,
                                            max_history_size=self.settings.module_llama_cpp_prompt_history_size,
                                            parent=ai_dialog)

    def get_prompt_manager(self):
        return self.prompt_manager

    @staticmethod
    def get_lexemes_path():
        return os.path.join(os.path.dirname(__file__), 'lexemes')

    def attach_dialog(self, ai_dialog):
        # Catch parent dialog's closing event to handle cleanup
        ai_dialog.dialog_closed.connect(self.parent_closed)
        # Handle request cancelled events gracefully
        ai_dialog.request_cancelled.connect(self.request_cancelled)
        # Connect update signal to the dialog for real-time updates
        self.update_signal.connect(ai_dialog.append_to_message)
        # Connect to the token usage update signal for monitoring usage
        self.update_usage_signal.connect(ai_dialog.update_usage)
        # Initialize and connect the prompt manager
        self.init_prompt_manager(ai_dialog)

    @asyncSlot()
    async def parent_closed(self):
        ai_dialog = self.sender()
        # Use try/except to handle potential asynchronous deletions of objects during callback processing.
        # Explicitly check that all connections are disconnected to prevent memory leaks.
        try:
            if ai_dialog:
                # Disconnect from the messages updating signal
                self.update_signal.disconnect(ai_dialog.append_to_message)
                # Disconnect from the token usage update signal
                self.update_usage_signal.disconnect(ai_dialog.update_usage)
                if hasattr(ai_dialog, 'dialog_closed'):
                    # Ensure handling when the parent window might be closed
                    ai_dialog.dialog_closed.disconnect(self.parent_closed)
        except RuntimeError as e:  # Object might be already deleted
            self.logger.warning(f'Error occurred during the closing process {e}')
        # Clear the prompt history when no longer needed
        PromptManager.reload()
        # Stop the generator thread before closing the dialog to avoid crashes
        await self.stop_generator()

    @asyncSlot()
    async def request_cancelled(self):
        # Stop the generator thread before closing the dialog to avoid crashes
        await self.stop_generator()

    @asyncSlot()
    async def request(self, user_prompt: list, request_msg_id: int, response_msg_id: int,
                      init_callback: Callable = None, finished_callback: Callable = None):
        """
        Builds a request to the selected LLM functionality.

        Args:
            user_prompt (list): A list of dictionaries that form a chat history, managed by a prompt manager.
            request_msg_id (int): The ID of the request message.
            response_msg_id (int): The ID of the response message, which is the actual reply to the requested message.
            init_callback (Callable): The initialization callback that is triggered when the model is ready to start
                                      inferencing.
            finished_callback (Callable): The callback that is triggered upon completion of the inference.

        Returns:
            None
        """

        if not asyncio.get_running_loop().is_running():
            return

        # Verify that the prompt manager is activated before proceeding
        if not hasattr(self, 'prompt_manager') or self.prompt_manager is None:
            return

        # Store callbacks for later use or cleanup
        self.init_callback = init_callback
        self.finished_callback = finished_callback

        # Initialize the model
        try:
            self.model_helper.init_model()
        except (RuntimeError, ValueError, Exception) as e:
            self.logger.error(f'{e}')
            # Complete the initialization process
            if self.init_callback and callable(self.init_callback):
                self.init_callback()
            # Prepare error messages for display in the UI
            outputs = self.lexemes.get("module_llama_cpp_model_exception", scope='common', error_msg=str(e))
            # Emit update message signal
            self.update_signal.emit(outputs, None, None, EnumMessageType.DEFAULT, EnumMessageStyle.ERROR)
            # Emit finished signal
            self.finished_callback(request_msg_id=request_msg_id, response_msg_id=response_msg_id,
                                   message_type=EnumMessageType.RESPONSE)
            return

        user_prompt_str = self.prompt_manager.get_history()
        input_tokens = self.model_helper.get_input_tokens(user_prompt_str)

        # Emit update usage signal to reflect new token counts
        self.update_usage_signal.emit(self.get_model_name(), len(input_tokens), 0, len(input_tokens), False)

        if not self.generator_task or self.generator_task.done():
            # self.generator_task = await asyncio.create_task(self.run_generator(search_options, input_tokens))
            # Initialize generator task asynchronously with the configured prompt and message ids
            self.generator_task = asyncio.ensure_future(
                self.run_generator(user_prompt, request_msg_id, response_msg_id))
            # Handle the completion of tasks with the appropriate callback
            if self.finished_callback and callable(self.finished_callback):
                self.generator_task.add_done_callback(
                    lambda _task, _request_msg_id=request_msg_id, _response_msg_id=response_msg_id:
                    self.finished_callback(request_msg_id=_request_msg_id, response_msg_id=_response_msg_id,
                                           message_type=EnumMessageType.RESPONSE))
            # Run the task asynchronously
            done, pending = await asyncio.wait([self.generator_task], return_when=asyncio.FIRST_COMPLETED)

            for task in done:
                if not task.cancelled():
                    exception = task.exception()
                    if exception:
                        self.logger.warning(f"Task raised an exception: {exception}")
                        # Prepare error messages for UI display if exceptions occur
                        outputs = self.lexemes.get('module_llama_cpp_task_exception', scope='common',
                                                   error_msg=str(exception))
                        # Emit update message signal
                        self.update_signal.emit(outputs, None, None, EnumMessageType.DEFAULT, EnumMessageStyle.ERROR)
                    else:
                        result = task.result()
                        self.logger.debug(f"Task completed with result: {result}")

    async def run_generator(self, user_prompt, request_msg_id, response_msg_id):

        try:
            generator = self.model_helper.init_generator(user_prompt, ModelHelper.search_options)
        except ModuleNotFoundError as e:
            self.logger.error(f'Cannot init model generator: {e}', exc_info=False)
            # Re-raise exceptions to indicate unresolved issues to the caller
            raise
        finally:
            # Mark the end of the initialization stage
            if self.init_callback and callable(self.init_callback):
                self.init_callback()

        try:
            while not self.generator_task.done():
                await self.async_generator(generator, request_msg_id, response_msg_id)
                await asyncio.sleep(0.01)  # Sleep briefly to allow the UI to update
        except StopAsyncIteration:
            self.logger.debug("Async generation completed")
        except asyncio.CancelledError:
            self.logger.info("Generation cancelled")
        except (WindowsError, RecursionError) as e:
            self.logger.error(f"Error occurred: {e}")
        except (SystemExit, Exception) as e:
            self.logger.error(f"Exception raised: {e}")
        finally:  # Mark the end of the generation process
            await self.stop_generator()

    async def async_generator(self, generator, request_msg_id, response_msg_id):
        # Retrieve outputs for this iteration to process or display
        outputs = await self.model_helper.generate_output(generator)
        if len(outputs) > 0:  # there might be a space symbol
            # Emit update message signal
            self.update_signal.emit(outputs, request_msg_id, response_msg_id,
                                    EnumMessageType.RESPONSE, EnumMessageStyle.DEFAULT)
            # Emit update usage signal
            self.update_usage_signal.emit(self.get_model_name(), 0, 1, 1, True)  # One token at the moment

    async def stop_generator(self):
        # Cancel asynchronous tasks cleanly when necessary.
        if self.generator_task and not self.generator_task.done():
            # Ensure to finish callback execution; do not remove:
            # > self.generator_task.remove_done_callback(self.finished_callback)
            self.generator_task.cancel()

    def extend_settings_dialog_fields_conf(self, tab_widget) -> list:
        # Configuration settings for Module llama.cpp
        tab_module_llama_cpp_config = QWidget()

        # Create the scroll area
        scroll_area = QScrollArea()
        scroll_area.setObjectName('settings_dialog_tab_module_llama_cpp_config')
        scroll_area.setWidgetResizable(True)

        # Define the layout for the Module llama.cpp configuration tab
        tab_module_llama_cpp_config_layout = QVBoxLayout(tab_module_llama_cpp_config)

        # Set the content widget inside the scroll area
        scroll_area.setWidget(tab_module_llama_cpp_config)

        tab_widget.addTab(scroll_area, self.lexemes.get('tab_module_llama_cpp_config'))

        return [
            # [Module llama.cpp config]
            # Label for the configuration block
            {"type": QLabel, "name": "settings_dialog_module_llama_cpp_config_label",
             "props": {"setProperty": ("class", "group-header-label")},
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('module_llama_cpp_config_label'), "style": {"bold": True},
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj)},
            # Label for the model path input field
            {"type": LabelWithHint, "kwargs": {
                "tooltip": ('module_llama_cpp_config_path_input_accessible_description',
                            self.lexemes.get('module_llama_cpp_config_path_input_accessible_description'))},
             "name": "settings_dialog_module_llama_cpp_config_path_label",
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('module_llama_cpp_config_path_label'),
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Input field for the model path
            {"type": FilePathLineEdit, "kwargs": {"settings": self.settings, "ext_filter": "%s (*.gguf)" %
                                                  # The only GGUF files are now supported
                                                  self.lexemes.get('module_llama_cpp_config_path_input_filter_text')},
             "name": "settings_dialog_module_llama_cpp_config_path:module_llama_cpp_model_path",
             "read_only": False, "max_length": 256,
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "placeholder_text": self.lexemes.get('module_llama_cpp_config_path_input_placeholder_text'),
             "accessible_description":
                 self.lexemes.get('module_llama_cpp_config_path_input_accessible_description')},
            # Vertical spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj)},
            # Label for the context window setting
            {"type": LabelWithHint, "kwargs": {
                "tooltip": ('module_llama_cpp_config_context_window_input_accessible_description',
                            self.lexemes.get('module_llama_cpp_config_context_window_input_accessible_description'))},
             "name": "settings_dialog_module_llama_cpp_config_context_window_label",
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('module_llama_cpp_config_context_window_label'),
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Input field for the context window setting
            {"type": QSpinBox, "props": {'setMinimum': 1, 'setMaximum': 65536},  # Update the highest range
             "size_policy": (QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred),
             "name": "settings_dialog_module_llama_cpp_config_context_window:"
                     "module_llama_cpp_context_window",
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "accessible_description":
                 self.lexemes.get('module_llama_cpp_config_context_window_input_accessible_description')},
            # Vertical spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj)},
            # Label for supported chat formats
            {"type": LabelWithHint, "kwargs": {
                "tooltip": ('module_llama_cpp_chat_formats_combo_accessible_description',
                            self.lexemes.get('module_llama_cpp_chat_formats_combo_accessible_description'))},
             "name": "settings_dialog_module_llama_cpp_chat_formats_label",
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('module_llama_cpp_chat_formats_label'),
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj)},
            # Dropdown for selecting a supported chat format
            {"type": EnumComboBox, "args": [sorted(LlmChatFormats, key=lambda member: member.is_default)],
             "name": "settings_dialog_module_llama_cpp_chat_formats_combo:"
                     "module_llama_cpp_chat_format",
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj),
             "placeholder_text": self.lexemes.get('module_llama_cpp_chat_formats_combo_placeholder_text'),
             "accessible_description":
                 self.lexemes.get('module_llama_cpp_chat_formats_combo_accessible_description')},
            # Horizontal line spacer
            {"type": HorizontalLineSpacer, "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj)},
            # Label for the system prompt text editor
            {"type": LabelWithHint, "kwargs": {
                "tooltip": ('module_llama_cpp_config_system_prompt_edit_accessible_description',
                            self.lexemes.get('module_llama_cpp_config_system_prompt_edit_accessible_description'))},
             "name": "settings_dialog_module_llama_cpp_config_system_prompt_label",
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('module_llama_cpp_config_system_prompt_label'),
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Text editor field for the system prompt
            {"type": QPlainTextEdit,
             "name": "settings_dialog_module_llama_cpp_system_prompt:"
                     "module_llama_cpp_system_prompt",
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "placeholder_text": self.lexemes.get('module_llama_cpp_config_system_prompt_edit_placeholder_text'),
             "accessible_description":
                 self.lexemes.get('module_llama_cpp_config_system_prompt_edit_accessible_description'),
             "text_lines": 7},
            # Vertical spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj)},
            # Label for the temperature slider
            {"type": LabelWithHint, "kwargs": {
                "tooltip": ('module_llama_cpp_config_response_temperature_input_accessible_description',
                            self.lexemes.get('module_llama_cpp_config_response_temperature_input_accessible_description'))},
             "name": "settings_dialog_module_llama_cpp_config_response_temperature_label",
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('module_llama_cpp_config_response_temperature_label',
                                      temperature=self.settings.module_llama_cpp_response_temperature),
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Slider to adjust the temperature setting
            {"type": QSlider, "args": [Qt.Orientation.Horizontal],
             "props": {'setFocusPolicy': Qt.FocusPolicy.StrongFocus, 'setTickPosition': QSlider.TickPosition.TicksAbove,
                       'setTickInterval': 5, 'setSingleStep': 5, 'setMinimum': 0, 'setMaximum': 100},
             "name": "settings_dialog_module_llama_cpp_config_response_temperature:"
                     "module_llama_cpp_response_temperature",  # Lexeme key : Object name
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj),
             "on_value_change": self.temperature_change_handler,
             "accessible_description":
                 self.lexemes.get('module_llama_cpp_config_response_temperature_input_accessible_description')},
            # Vertical spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj)},
            # Label for the maximum token count setting
            {"type": LabelWithHint, "kwargs": {
                "tooltip": ('module_llama_cpp_config_response_max_tokens_input_accessible_description',
                            self.lexemes.get('module_llama_cpp_config_response_max_tokens_input_accessible_description'))},
             "name": "settings_dialog_module_llama_cpp_config_response_max_tokens_label",
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('module_llama_cpp_config_response_max_tokens_label'),
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Input field for the maximum token count setting
            {"type": QSpinBox, "props": {'setMinimum': 0, 'setMaximum': 65536},  # Update the highest range
             "size_policy": (QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred),
             "name": "settings_dialog_module_llama_cpp_config_response_max_tokens:"
                     "module_llama_cpp_response_max_tokens",
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "accessible_description":
                 self.lexemes.get('module_llama_cpp_config_response_max_tokens_input_accessible_description')},
            # Horizontal line spacer
            {"type": HorizontalLineSpacer, "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj)},
            # Label for the prompt history maximum capacity settings
            {"type": LabelWithHint, "kwargs": {
                "tooltip": ('module_llama_cpp_config_prompt_history_size_input_accessible_description',
                            self.lexemes.get('module_llama_cpp_config_prompt_history_size_input_accessible_description'))},
             "name": "settings_dialog_module_llama_cpp_config_prompt_history_size_label",
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('module_llama_cpp_config_prompt_history_size_label'),
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Input field for the prompt history maximum capacity setting
            {"type": QSpinBox, "props": {'setMinimum': 0, 'setMaximum': 65536},  # Update the highest range
             "size_policy": (QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred),
             "name": "settings_dialog_module_llama_cpp_config_prompt_history_size:"
                     "module_llama_cpp_prompt_history_size",
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "accessible_description":
                 self.lexemes.get('module_llama_cpp_config_prompt_history_size_input_accessible_description')},
            # Spacer to align elements at the top of the layout
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding),
             "callback": lambda obj: tab_module_llama_cpp_config_layout.addWidget(obj)},
        ]

    def get_model_name(self):
        # Retrieve the LLM model name
        return f'({self.model_helper.get_model_name()})'

    @staticmethod
    def extend_settings_create_property(extend_func: Callable):
        if callable(extend_func):
            extend_func("module_llama_cpp_model_path", str, "")
            extend_func("module_llama_cpp_context_window", int, 2048)
            extend_func("module_llama_cpp_chat_format", str, str(LlmChatFormats.default()))
            extend_func("module_llama_cpp_system_prompt", str, "")
            extend_func("module_llama_cpp_response_temperature", int, 20)
            extend_func("module_llama_cpp_response_max_tokens", int, 0)
            extend_func("module_llama_cpp_prompt_history_size", int, 0)

    def settings_update_handler(self, data) -> None:
        """
        Perform actions upon settings change. This function handles data represented as a dictionary,
        where each key is the setting name and the corresponding value is the setting's new value.
        Note: This operation can be resource-intensive.

        Args:
            data (dict): Dictionary containing settings changes, e.g., {"module_llama_cpp_model_path": "..."}

        Returns:
            None
        """

        AppConfig().logger.debug('Settings update handler is in use "%s"' % data)

        options = [
            'module_llama_cpp_model_path',
            'module_llama_cpp_context_window',
            'module_llama_cpp_chat_format',
            'module_llama_cpp_response_temperature',
            'module_llama_cpp_response_max_tokens',
            # 'module_llama_cpp_system_prompt',
            # 'module_llama_cpp_prompt_history_size',
        ]
        if any(option in data for option in options) or 'ai_config_inference_module' in data:
            # Apply updated values or revert to settings if necessary
            model_path = self.settings.module_llama_cpp_model_path
            if 'module_llama_cpp_model_path' in data:
                model_path = data['module_llama_cpp_model_path']

            context_window = self.settings.module_llama_cpp_context_window
            if 'module_llama_cpp_context_window' in data:
                context_window = data['module_llama_cpp_context_window']

            chat_format = self.settings.module_llama_cpp_chat_format
            if 'module_llama_cpp_chat_format' in data:
                chat_format = data['module_llama_cpp_chat_format']

            # Reload the model helper
            ModelHelper.reload()

            search_options = self.get_search_options()

            # The model will be reloaded for the next session
            self.model_helper = ModelHelper(model_path=model_path, n_ctx=context_window, chat_format=chat_format,
                                            search_options=search_options)

    def temperature_change_handler(self, source_object, source_widget):
        if source_object.objectName() and source_widget:
            # Parse the object name to handle combinations of lexeme and setting keys
            _lexeme_key, setting_name = self.settings.settings_helper.parse_object_name(source_object.objectName())
            if setting_name == 'module_llama_cpp_response_temperature':
                # Fetch the latest value from the object, considering that settings might not be fully updated yet.
                temperature_int = source_object.value()  # source_object type: QSlider
                temperature_float = self.model_helper.convert_temperature(temperature_int)
                temperature_label = source_widget.findChild(
                    QLabel,
                    'settings_dialog_module_llama_cpp_config_response_temperature_label'
                )
                if temperature_label:
                    temperature_label.setText(
                        self.lexemes.get('module_llama_cpp_config_response_temperature_label',
                                         temperature=temperature_float))



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Extends AI Assistant Dialog Class by formatting and displaying messages from and to the user.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtWidgets import QStyle, QSizePolicy, QLabel, QPushButton
from PySide6.QtGui import QTextDocument, QColor
from PySide6.QtCore import Qt

from . import Lexemes
from . import ThemeHelper
from . import ClipboardHelper
from . import TooltipHelper

import logging


class AIMessageLabel(QLabel):

    button: QPushButton = None

    def __init__(self, parent=None, text=None, color=None, bg_color=None, settings=None):
        super(AIMessageLabel, self).__init__(parent)

        self.parent = parent

        if self.parent and hasattr(self.parent, 'font'):
            # Apply font from the main window to the element
            self.setFont(self.parent.font())

        # The class may be in use within settings itself
        self.settings = settings if settings else self.parent.settings  # type: ignore

        self.logger = logging.getLogger('ai_message_label')

        # Load lexemes for selected language and scope
        self.lexemes = Lexemes(self.settings.app_language, default_scope='ai_assistant')

        self.theme_helper = ThemeHelper()

        # Set the text
        if text:
            self.setText(text)

        # Assign colors to QLabel for styling
        self.color = color if color else 'palette(text)'
        self.bg_color = bg_color if bg_color else 'palette(base)'

        self.init_ui()

    def init_ui(self):
        # Set QLabel attributes (e.g., text, font, and interaction flags)
        self.setWordWrap(True)
        self.setStyleSheet(""" QLabel {
            border-radius: 5px;
            margin: 5px 0;
            padding: 5px;
            color: %s;
            background-color: %s;
        } """ % (self.color, self.bg_color))

        palette = self.palette()
        palette.setColor(self.foregroundRole(), QColor(self.color))
        palette.setColor(self.backgroundRole(), QColor(self.bg_color))

        self.setPalette(palette)
        self.setAutoFillBackground(True)
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)

        self.button = QPushButton(self)

        # copy_icon = self.style().standardIcon(QStyle.StandardPixmap.SP_DirIcon)  # Set an icon using standard icons
        copy_icon = self.theme_helper.get_icon(
            theme_icon='copy.svg',
            color=QColor(self.theme_helper.get_color('ai_assistant_message_icon_color', css_format=True))
        )
        self.button.setIcon(copy_icon)

        self.button.clicked.connect(self.copy_content)
        self.button.setCursor(Qt.CursorShape.PointingHandCursor)
        self.button.setFixedSize(20, 20)
        self.button.setStyleSheet("QPushButton { border: none; padding: 0; }")
        self.button.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.button.setToolTip(self.lexemes.get('dialog_message_copy_tooltip'))

        # Ensure the text doesn't overlap the button
        button_size = self.button.sizeHint()
        self.setContentsMargins(0, 0, button_size.width(), 0)

        # Enable text selection in QLabel using the mouse
        self.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)

    def resizeEvent(self, event):
        button_size = self.button.sizeHint()
        frame_width = self.style().pixelMetric(QStyle.PixelMetric.PM_DefaultFrameWidth)
        self.button.move(self.rect().right() - frame_width - button_size.width(),
                         (self.rect().top() + button_size.height() + 2) // 2)
        super(AIMessageLabel, self).resizeEvent(event)

    def copy_content(self):
        # Copy text to the clipboard
        ClipboardHelper.set_text(self.get_text_to_copy())
        # Show tooltip
        TooltipHelper.show_tooltip(widget=self.button, text=self.lexemes.get('dialog_message_copied_tooltip'))

    def get_text_to_copy(self):
        # Check if the result needs to be converted from HTML to raw text.
        if getattr(self.settings, 'ai_config_convert_to_md', False):
            # Create a QTextDocument from the label's HTML content.
            doc = QTextDocument()
            doc.setHtml(self.text())

            # Convert the HTML content to raw text.
            return doc.toRawText()

        return self.text()



# Swedish lexemes toolbar.py
lexemes = {
    "actions_label_new_document": "Nytt dokument",
    "actions_accessible_name_new_document": "Nytt dokument",
    "actions_label_open": "Öppna",
    "actions_accessible_name_open": "Öppna dokument",
    "actions_label_save": "Spara",
    "actions_accessible_name_save": "Spara dokument",
    "actions_label_save_as": "Spara som",
    "actions_accessible_name_save_as": "Spara som dokumentkopia",
    "actions_label_sync": "Synkronisera",
    "actions_accessible_name_sync": "Synkronisera dokument med extern lagring",
    "actions_label_edit": "Redigera",
    "actions_accessible_name_edit": "Redigera fil",
    "actions_label_source": "Källkod",
    "actions_accessible_name_source": "Källkod",
    "actions_label_encrypt": "Kryptera",
    "actions_accessible_name_encrypt": "Kryptera fil",
    "actions_label_decrypt": "Dekryptera",
    "actions_accessible_name_decrypt": "Dekryptera fil",
    "actions_label_bold": "Fet",
    "actions_accessible_name_bold": "Fet textformat",
    "actions_label_italic": "Kursiv",
    "actions_accessible_name_italic": "Kursiv textformat",
    "actions_label_underline": "Understrykning",
    "actions_accessible_name_underline": "Understrykningsformat",
    "actions_label_strikethrough": "Genomstrykning",
    "actions_accessible_name_strikethrough": "Genomstrykningsformat",
    "actions_label_blockquote": "Blockcitat",
    "actions_accessible_name_blockquote": "Blockcitatformat",
    "actions_label_color_picker": "Färgväljare",
    "actions_accessible_name_color_picker": "Välj textfärg",
    "actions_label_ai_assistant": "AI-assistent",
    "actions_accessible_name_ai_assistant": "Starta AI-assistent",
    "actions_label_about": "Om",
    "actions_accessible_name_about": "Om appen",
    "actions_label_exit": "Avsluta",
    "actions_accessible_name_exit": "Avsluta appen",
    "actions_label_settings": "Inställningar",
    "actions_accessible_name_settings": "Appinställningar",

    "search_input_label": "Sök",
    "search_input_placeholder_text": "Sök text i dokumentet",
    "search_input_accessible_description": "Sökfält",
    "search_buttons_label_clear": "Rensa sökning",
    "search_buttons_accessible_name_clear": "Rensa sökfält",
    "search_buttons_label_prev": "Föregående sökning",
    "search_buttons_accessible_name_prev": "Föregående sökresultat",
    "search_buttons_label_next": "Nästa sökning",
    "search_buttons_accessible_name_next": "Nästa sökresultat",
    "search_case_sensitive_accessible_name": "Skiftlägeskänslig",
    "search_case_sensitive_tooltip": "Skiftlägeskänslig sökning"
}



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Part of the 'OpenAI API' module.
- Functionality: Facilitates initialization and management of OpenAI API requests and responses.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import QUrl, QByteArray
from PySide6.QtNetwork import QNetworkRequest

import json
import logging

from threading import Lock


class ApiHelper:

    _instance = None  # Singleton instance
    _lock = Lock()

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            with cls._lock:
                if not cls._instance:
                    # Create the instance if it doesn't exist
                    cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, parent=None):
        # Prevent re-initialization if the instance is already set up.
        if hasattr(self, 'logger'):
            return

        # Use a lock to ensure initialization is thread-safe and atomic.
        with self._lock:
            # Double-check to prevent race conditions during initialization.
            if hasattr(self, 'logger'):
                return

            super(ApiHelper, self).__init__()

            self.parent = parent

            self.logger = logging.getLogger('openai_api_helper')

    def init_request(self, api_url, api_key) -> QNetworkRequest:

        url = QUrl(api_url)  # API endpoint
        request = QNetworkRequest(url)

        # Set Authorization header
        request.setRawHeader(b"Authorization", bytes("Bearer " + api_key, encoding="utf-8"))

        # Set Content-Type header
        request.setHeader(QNetworkRequest.KnownHeaders.ContentTypeHeader, "application/json")

        # Set request type to POST
        request.setRawHeader(b"Custom-Request", b"POST")

        return request

    def init_request_params(self, prompt_messages, api_model,
                            # Optional arguments
                            options: dict = None) -> QByteArray:

        # New completions
        post_params = {
            "model": api_model,
            "messages": prompt_messages,
            "temperature": 0.2,
            "top_p": 1,
            "n": 1,
            "stream": False,
        }
        # If response max tokens set
        if 'max_tokens' in options:
            post_params.update({"max_tokens": options['max_tokens']})
        # Other params to override
        if 'temperature' in options:
            post_params.update({"temperature": options['temperature']})
        if 'top_p' in options:
            post_params.update({"top_p": options['top_p']})
        if 'n' in options:
            post_params.update({"n": options['n']})
        if 'stream' in options:
            post_params.update({"stream": bool(options['stream'])})

        json_post_params = json.dumps(post_params)

        # Set request data
        request_data = QByteArray(json_post_params.encode("utf-8"))

        return request_data

    # def get_prompt_from_text(self, text):
    #    prompt = f'{self.chat_template.format(input=text)}'
    #    return prompt

    @staticmethod
    def convert_temperature(temperature: int = 0):
        """ Convert the integer value of temperature to a float. """
        return temperature / 100



# French lexemes settings_dialog.py
lexemes = {
    # Paramètres
    "window_title": "Paramètres",

    "button_close": "Fermer",

    "tab_general": "Général",
    "tab_editor_config": "Éditeur",
    "tab_viewer_config": "Visualiseur",
    "tab_ai_config": "Config IA",

    "general_app_config_label": "Configuration de l'application",
    "general_app_language_label": "Langue",
    "general_app_language_combo_placeholder_text": "Choisir une langue",
    "general_app_language_combo_accessible_description": "Langue de l'interface de l'application",
    "general_app_theme_label": "Thème",
    "general_app_theme_combo_placeholder_text": "Choisir un thème",
    "general_app_theme_combo_accessible_description": "Thème de l'interface de l'application",
    "general_app_main_menu_label": "Menu Principal",
    "general_app_main_menu_checkbox": "Afficher le menu principal",
    "general_app_main_menu_checkbox_accessible_description": "Afficher le menu déroulant principal de l'application",
    "general_app_font_size_label": "Taille de police : {size}pt",
    "general_app_font_size_slider_accessible_description": "Ajuster la taille de la police globale de l'application",

    "general_statusbar_label": "Barre de statut",
    "general_statusbar_show_global_cursor_position_checkbox": "Afficher la position globale du curseur",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Afficher la position globale du curseur dans la barre de statut",

    "editor_config_label": "Configuration de l'éditeur",
    "editor_config_show_line_numbers_checkbox": "Afficher les numéros de ligne",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Afficher les numéros de ligne dans l'éditeur",

    "viewer_config_label": "Configuration du visualiseur",
    "viewer_config_process_emojis_checkbox": "Convertir les emojis de texte en graphiques",
    "viewer_config_process_emojis_checkbox_accessible_description":
        "Convertir les emojis de texte en représentations graphiques",
    "viewer_config_highlight_todos_checkbox": "Mettre en évidence les TODO",
    "viewer_config_highlight_todos_checkbox_accessible_description": "Souligner les balises TODO dans le texte",
    "viewer_config_open_link_confirmation_checkbox": "Demander confirmation avant d'ouvrir les liens",
    "viewer_config_open_link_confirmation_checkbox_accessible_description":
        "Demander une confirmation avant d'ouvrir des liens",
    "viewer_config_save_resources_checkbox": "Enregistrer automatiquement les images externes sur le disque",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Enregistre automatiquement des copies des images externes sur le disque pour un accès hors ligne.",

    "ai_config_inference_module_label": "Module d'inférence",
    "ai_config_inference_module_names_combo_label": "Module d'inférence actif",
    "ai_config_inference_module_names_combo_placeholder_text": "Choisir un module",
    "ai_config_inference_module_names_combo_accessible_description":
        "Sélectionnez parmi les modules d'inférence IA disponibles pour fonctionner avec l'Assistant AI.\n"
        "Les options incluent des modèles de langage de grande taille (LLM) avec traitement en temps réel,\n"
        "ou des fonctionnalités basées sur l'API.",

    "ai_config_base_label": "Paramètres de base",
    "ai_config_multi_turn_dialogue_checkbox": "Dialogue multi-tours avec mémoire conversationnelle",
    "ai_config_multi_turn_dialogue_checkbox_accessible_description":
        "Activez un dialogue multi-tours qui conserve la dernière invite pour une mémoire conversationnelle.\n"
        "Lorsqu'il est désactivé, seuls le nouveau message et l'invite système influencent la réponse.",
    "ai_config_convert_to_md_checkbox": "Convertir le résultat en Markdown",
    "ai_config_convert_to_md_checkbox_accessible_description":
        "Convertissez le message de sortie au format Markdown.",
}



# Chinese lexemes settings_dialog.py
lexemes = {
    "tab_module_llama_cpp_config": "模块 llama.cpp",

    "module_llama_cpp_config_label": "模块 llama.cpp",
    "module_llama_cpp_config_path_label": "模型位置",
    "module_llama_cpp_config_path_input_placeholder_text": "选择或输入模型路径",
    "module_llama_cpp_config_path_input_accessible_description":
        "一个带选择器的输入字段，用于指定本地模型的路径。支持GGUF格式的模型，\n"
        "一种二进制文件格式，专为存储用于GGML及基于GGML执行器的模型而优化。",
    "module_llama_cpp_config_path_input_filter_text": "GGUF 文件",

    "module_llama_cpp_config_context_window_label": "上下文窗口大小",
    "module_llama_cpp_config_context_window_input_accessible_description":
        "设置模型生成响应时考虑的令牌数量。控制使用多少先前上下文。",

    "module_llama_cpp_chat_formats_label": "聊天格式",
    "module_llama_cpp_chat_formats_combo_placeholder_text": "选择聊天格式",
    "module_llama_cpp_chat_formats_combo_accessible_description":
        "用于选择模型对话格式的下拉菜单。",

    "module_llama_cpp_config_system_prompt_label": "系统提示",
    "module_llama_cpp_config_system_prompt_edit_placeholder_text": "输入系统提示文本",
    "module_llama_cpp_config_system_prompt_edit_accessible_description":
        "用于输入指导模型响应的系统提示的文本字段。",

    "module_llama_cpp_config_response_temperature_label": "响应温度：{temperature}",
    "module_llama_cpp_config_response_temperature_input_accessible_description":
        "调整模型响应的随机性。较高的值产生更多变化的输出，\n"
        "而较低的值结果更可预测的响应。",

    "module_llama_cpp_config_response_max_tokens_label": "每次响应的最大令牌数",
    "module_llama_cpp_config_response_max_tokens_input_accessible_description":
        "限制模型响应中的令牌数量，直至实际的上下文窗口限制。\n"
        "零值假设上下文窗口的容量。",

    "module_llama_cpp_config_prompt_history_size_label": "提示历史大小",
    "module_llama_cpp_config_prompt_history_size_input_accessible_description":
        "控制系统保留供参考的提示历史记录中的条目数量。\n"
        "零值允许无限条目。"
}



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Part of the 'Module llama.cpp' module.
- Functionality: Manages prompt formatting and maintains a history of prompts.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import QObject

import logging

from threading import Lock
from typing import Union

from ...ui.ai_assistant.ai_assistant import EnumMessageType


class PromptManager(QObject):

    prompt_template = None
    prompt_multi_template = None

    supported_roles = ['system', 'user', 'assistant']  # 'tool', 'function'

    _instance = None  # Singleton instance
    _lock = Lock()

    history: list = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            with cls._lock:
                if not cls._instance:
                    # Create the instance if it doesn't exist
                    cls._instance = super().__new__(cls)
        return cls._instance

    @classmethod
    def reload(cls, *args, **kwargs):
        """
        Reinitialize the singleton instance to allow controlled re-creation.
        """
        with cls._lock:
            # Create a new instance
            cls._instance = super().__new__(cls, *args, **kwargs)

    def __init__(self, system_prompt=None, max_history_size=None, parent=None):

        # Prevent re-initialization if the instance is already set up.
        if hasattr(self, 'logger'):
            return

        # Use a lock to ensure initialization is thread-safe and atomic.
        with self._lock:
            # Double-check to prevent race conditions during initialization.
            if hasattr(self, 'logger'):
                return

            super().__init__(parent)

            self.logger = logging.getLogger('llama_cpp_prompt_manager')

            self.system_input = system_prompt
            self.max_history_size = max_history_size

            self.init_history()

            if parent and hasattr(parent, 'message_added'):
                # Catch new message added signal
                parent.message_added.connect(self.add_message)

    def init_history(self):
        self.logger.info('Initializing prompt history')
        # Start from new
        self.history = []
        # Init system prompt if set
        if self.system_input:
            self.add_system(self.system_input)

    def get_prompt_message(self, role: str, content: str) -> Union[dict, None]:
        if role not in PromptManager.supported_roles:
            self.logger.warning(f'Role "{role}" is not supported yet')
            return None
        return {'role': role, 'content': content}

    def get_prompt(self, multi_turn=True):
        if multi_turn:
            prompt = self.format_history()
        else:
            prompt = [
                self.get_prompt_message(role='system', content=self.system_input),
            ]
            last_user_message = self.find_last_message_by_role('user')
            if last_user_message:
                prompt.append(last_user_message)
            else:
                self.logger.warning('No user message has found for prompt')
                return None
        return prompt

    def add_message(self, message_text, request_msg_id, response_msg_id, message_type: EnumMessageType):
        self.logger.debug(f'Add message: {message_text}, {response_msg_id}, {message_type}')
        if message_type == EnumMessageType.USER_INPUT:
            self.add_request(message_text, request_msg_id)
        elif message_type == EnumMessageType.RESPONSE:
            self.add_response(message_text, response_msg_id, request_msg_id)
        else:
            # Default or info message; not for prompt
            pass

    def add_system(self, system_instruct, msg_id=None):
        """
        Add a new request to the history, initially without a response.

        Args:
            system_instruct (str): The text of the request.
            msg_id (int, optional): The unique message ID of the request.
        """
        self.update_or_add_entry(msg_id=msg_id,
                                 data=self.get_prompt_message(role='system', content=system_instruct))

    def add_request(self, request_text, request_msg_id):
        """
        Add a new request to the history, initially without a response.

        Args:
            request_text (str): The text of the request.
            request_msg_id (int): The unique message ID of the request.
        """
        self.update_or_add_entry(msg_id=request_msg_id,
                                 data=self.get_prompt_message(role='user', content=request_text))

    def add_response(self, response_text, response_msg_id, request_msg_id):
        """
        Update the corresponding request with a response, identified by request message ID.

        Args:
            response_text (str): The text of the response.
            response_msg_id (int): The message ID of the response.
            request_msg_id (int): The request message ID to which this response corresponds.
        """
        self.update_or_add_entry(msg_id=response_msg_id,
                                 data=self.get_prompt_message(role='assistant', content=response_text))

    def update_or_add_entry(self, msg_id, data):
        """
        Update the entry with the specified ID, or add a new entry if ID does not exist.

        Args:
        - msg_id (int): The ID to search for in the messages.
        - data (dict): The new data to replace the old 'data' dictionary or to add if ID is new.

        Returns:
        - None; the list is modified in place.
        """
        for message in self.history:
            if message['id'] == msg_id:
                message['data'] = data
                return

        # If no entry with the specified ID is found, add a new one
        self.history.append({'id': msg_id, 'data': data})

        # Limit history records size (if applicable)
        self.limit_history_size()

    def format_history(self):
        """
        Format the stored history of requests and responses into a text format.

        Returns:
            str: A string representing the formatted history.
        """
        formatted_history = []  # In the API itself it looks like a: history = {"messages": [...]}

        for message in self.history:
            data = message['data']
            formatted_history.append(data)
        return formatted_history

    def get_history(self) -> str:
        # Get history object
        history = self.format_history()

        # Format each entry and combine them into a text
        result = "\n\n".join(f"**{entry['role']}**:\n{entry['content']}" for entry in history)

        return result

    def limit_history_size(self):
        """
        Ensures the history does not exceed the maximum specified size.
        If a system prompt is set, it retains the system prompt at the beginning of the history,
        and removes the oldest records to maintain the history within the specified size limit.
        """

        if self.max_history_size:  # Ensure max_history_size is set before enforcing limits
            # Calculate the additional index offset if a system prompt is included
            system_prompt_shift = 1 if self.system_input else 0
            # Calculate the total allowed history size including the system prompt
            allowed_history_size = self.max_history_size + system_prompt_shift

            # Remove entries until the history size is within the allowed limit
            while len(self.history) > allowed_history_size:
                # Remove the oldest record, keeping the system prompt intact if present
                self.history.pop(system_prompt_shift)  # Pop from the first non-prompt position

    def find_last_message_by_role(self, role: str) -> Union[dict, None]:
        if role not in PromptManager.supported_roles:
            self.logger.warning(f'Role "{role}" is not supported yet')
            return None
        for message in reversed(self.history):
            if 'data' in message and 'role' in message['data'] and message['data']['role'] == role:
                # Return the last role's message
                return message['data']
        return None
